# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import pydantic

import spectre_core.events
import spectre_core.flowgraphs

from ._validators import (
    skip_validator,
    validate_window_size,
    validate_one_of,
    validate_non_overlapping_steps,
    validate_num_samples_per_step,
    validate_nyquist_criterion,
    validate_num_steps_per_sweep,
)
from ._sdrplay_validators import (
    validate_bandwidth,
    validate_center_frequency,
    validate_if_gain,
    validate_low_if_sample_rate,
    validate_rf_gain,
    validate_sample_rate,
    validate_constant_lna_state,
)


def _get_rf_gains(center_frequency: float) -> list[int]:
    # Assuming high z is not enabled.
    if center_frequency <= 60e6:
        return [0, -6, -12, -18, -37, -42, -61]
    elif center_frequency <= 420e6:
        return [0, -6, -12, -18, -20, -26, -32, -38, -57, -62]
    elif center_frequency <= 1e9:
        return [0, -7, -13, -19, -20, -27, -33, -39, -45, -64]
    elif center_frequency <= 2e9:
        return [0, -6, -12, -20, -26, -32, -38, -43, -62]
    else:
        return []


class RSPduoFixedCenterFrequency(
    spectre_core.flowgraphs.RSPduoFixedCenterFrequencyModel,
    spectre_core.events.FixedCenterFrequencyModel,
):
    @pydantic.model_validator(mode="after")
    def validator(self, info: pydantic.ValidationInfo):
        if skip_validator(info):
            return self
        validate_one_of(
            self.antenna_port,
            [
                spectre_core.flowgraphs.RSPduoPort.TUNER_1,
                spectre_core.flowgraphs.RSPduoPort.TUNER_2,
            ],
            name="antenna_port",
        )
        validate_nyquist_criterion(self.sample_rate, self.bandwidth)
        validate_center_frequency(self.center_frequency)
        validate_window_size(self.window_size)
        validate_sample_rate(self.sample_rate)
        validate_bandwidth(self.bandwidth)
        validate_if_gain(self.if_gain)
        validate_low_if_sample_rate(self.sample_rate)
        validate_rf_gain(self.rf_gain, _get_rf_gains(self.center_frequency))
        return self


class RSPduoSweptCenterFrequency(
    spectre_core.flowgraphs.RSPduoSweptCenterFrequencyModel,
    spectre_core.events.SweptCenterFrequencyModel,
):
    @pydantic.model_validator(mode="after")
    def validator(self, info: pydantic.ValidationInfo):
        if skip_validator(info):
            return self
        validate_one_of(
            self.antenna_port,
            [
                spectre_core.flowgraphs.RSPduoPort.TUNER_1,
                spectre_core.flowgraphs.RSPduoPort.TUNER_2,
            ],
            name="antenna_port",
        )
        validate_center_frequency(self.min_frequency)
        validate_center_frequency(self.max_frequency)
        validate_window_size(self.window_size)
        validate_sample_rate(self.sample_rate)
        validate_bandwidth(self.bandwidth)
        validate_if_gain(self.if_gain)
        validate_low_if_sample_rate(self.sample_rate)
        validate_non_overlapping_steps(self.frequency_step, self.sample_rate)
        validate_num_samples_per_step(self.window_size, self.samples_per_step)
        validate_num_steps_per_sweep(
            self.min_frequency, self.max_frequency, self.frequency_step
        )
        validate_constant_lna_state(
            self.min_frequency, self.max_frequency, _get_rf_gains
        )
        return self



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import subprocess
import shutil
import gzip
import datetime
import enum

import spectre_core.config


class CallistoInstrumentCode(enum.Enum):
    """e-Callisto network station codes."""

    ALASKA_ANCHORAGE = "ALASKA-ANCHORAGE"
    ALASKA_COHOE = "ALASKA-COHOE"
    ALASKA_HAARP = "ALASKA-HAARP"
    ALGERIA_CRAAG = "ALGERIA-CRAAG"
    ALMATY = "ALMATY"
    AUSTRIA_KRUMBACH = "AUSTRIA-Krumbach"
    AUSTRIA_MICHELBACH = "AUSTRIA-MICHELBACH"
    AUSTRIA_OE3FLB = "AUSTRIA-OE3FLB"
    AUSTRIA_UNIGRAZ = "AUSTRIA-UNIGRAZ"
    AUSTRALIA_ASSA = "Australia-ASSA"
    BIR = "BIR"
    CROATIA_VISNJAN = "Croatia-Visnjan"
    DENMARK = "DENMARK"
    EGYPT_ALEXANDRIA = "EGYPT-Alexandria"
    EGYPT_SPACEAGENCY = "EGYPT-SpaceAgency"
    FINLAND_SIUNTIO = "FINLAND-Siuntio"
    FINLAND_KEMPELE = "Finland-Kempele"
    GERMANY_DLR = "GERMANY-DLR"
    GLASGOW = "GLASGOW"
    GREENLAND = "GREENLAND"
    HUMAIN = "HUMAIN"
    HURBANOVO = "HURBANOVO"
    INDIA_GAURI = "INDIA-GAURI"
    INDIA_OOTY = "INDIA-OOTY"
    INDIA_UDAIPUR = "INDIA-UDAIPUR"
    JAPAN_IBARAKI = "JAPAN-IBARAKI"
    KASI = "KASI"
    MEXART = "MEXART"
    MEXICO_FCFM_UANL = "MEXICO-FCFM-UANL"
    MEXICO_LANCE_A = "MEXICO-LANCE-A"
    MEXICO_LANCE_B = "MEXICO-LANCE-B"
    MONGOLIA_UB = "MONGOLIA-UB"
    MRO = "MRO"
    MRT3 = "MRT3"
    MALAYSIA_BANTING = "Malaysia-Banting"
    NORWAY_EGERSUND = "NORWAY-EGERSUND"
    NORWAY_NY_AALESUND = "NORWAY-NY-AALESUND"
    NORWAY_RANDABERG = "NORWAY-RANDABERG"
    POLAND_GROTNIKI = "POLAND-Grotniki"
    ROMANIA = "ROMANIA"
    ROSWELL_NM = "ROSWELL-NM"
    SPAIN_PERALEJOS = "SPAIN-PERALEJOS"
    SSRT = "SSRT"
    SWISS_HB9SCT = "SWISS-HB9SCT"
    SWISS_HEITERSWIL = "SWISS-HEITERSWIL"
    SWISS_IRSOL = "SWISS-IRSOL"
    SWISS_LANDSCHLACHT = "SWISS-Landschlacht"
    SWISS_MUHEN = "SWISS-MUHEN"
    TRIEST = "TRIEST"
    TURKEY = "TURKEY"
    UNAM = "UNAM"
    URUGUAY = "URUGUAY"
    USA_BOSTON = "USA-BOSTON"


def _get_batch_name(station: str, date: str, time: str, code: str) -> str:
    """
    Create a standardised batch file name for a Spectre batch file.

    :param station: Station name.
    :param date: Observation date in 'YYYYMMDD' format.
    :param time: Observation time in 'HHMMSS' format.
    :param code: Numeric instrument code.
    :return: Formatted batch file name.
    """
    dt = datetime.datetime.strptime(f"{date}T{time}", "%Y%m%dT%H%M%S")
    formatted_time = dt.strftime(spectre_core.config.TimeFormat.DATETIME)
    return f"{formatted_time}_callisto-{station.lower()}-{code}.fits"


def _get_batch_components(gz_path: str) -> list[str]:
    """
    Extract station, date, time, and instrument code from the `.fit.gz` file name.

    :param gz_path: Path to the `.fit.gz` file.
    :return: List of [station, date, time, instrument_code].
    :raises ValueError: If file format is invalid.
    """
    file_name = os.path.basename(gz_path)
    if not file_name.endswith(".fit.gz"):
        raise ValueError(f"Invalid file extension: {file_name}. Expected .fit.gz")
    file_base_name = file_name.rstrip(".fit.gz")
    parts = file_base_name.split("_")
    if len(parts) != 4:
        raise ValueError(
            "Invalid file name format. Expected '[station]_[date]_[time]_[code].fit.gz'"
        )
    return parts


def _get_batch_path(gz_path: str) -> str:
    """
    Generate the full path for the Spectre batch file.

    :param gz_path: Path to the raw `.fit.gz` file.
    :return: Path to the corresponding batch file.
    """
    station, date, time, code = _get_batch_components(gz_path)
    batch_name = _get_batch_name(station, date, time, code)
    batch_start_time = batch_name.split("_")[0]
    dt = datetime.datetime.strptime(
        batch_start_time, spectre_core.config.TimeFormat.DATETIME
    )
    batch_dir = spectre_core.config.paths.get_batches_dir_path(
        year=dt.year, month=dt.month, day=dt.day
    )
    os.makedirs(batch_dir, exist_ok=True)
    return os.path.join(batch_dir, batch_name)


def _unzip_file_to_batches(gz_path: str) -> str:
    """
    Decompress a `.fit.gz` file and save it as a `.fits` batch file.

    :param gz_path: Path to the `.fit.gz` file.
    :return: The file path of the newly created batch file, as absolute paths within the container's file system.
    """
    fits_path = _get_batch_path(gz_path)
    with gzip.open(gz_path, "rb") as f_in, open(fits_path, "wb") as f_out:
        shutil.copyfileobj(f_in, f_out)
        return f_out.name


def _unzip_to_batches(tmp_dir: str) -> list[str]:
    """
    Decompress all `.gz` files in a temporary directory and save them as Spectre
    batch files.

    :param tmp_dir: Path to the temporary directory containing `.gz` files.
    :return: A list of file names of all newly created batch files, as absolute paths within the container's file system.
    """
    batch_file_names = []
    for entry in os.scandir(tmp_dir):
        if entry.is_file() and entry.name.endswith(".gz"):
            batch_file_names.append(_unzip_file_to_batches(entry.path))
            os.remove(entry.path)
    shutil.rmtree(tmp_dir)
    return batch_file_names


def _wget_callisto_data(
    instrument_code: str, year: int, month: int, day: int, tmp_dir: str
) -> None:
    """
    Download raw `.fit.gz` files from the e-Callisto network using `wget`.

    :param instrument_code: Instrument code for filtering files.
    :param year: Observation year.
    :param month: Observation month.
    :param day: Observation day.
    :param tmp_dir: Path to the temporary directory for downloads.
    """
    date_str = f"{year:04d}/{month:02d}/{day:02d}"
    base_url = f"http://soleil.i4ds.ch/solarradio/data/2002-20yy_Callisto/{date_str}/"
    command = [
        "wget",
        "-r",
        "-l1",
        "-nd",
        "-np",
        "-R",
        ".tmp",
        "-A",
        f"{instrument_code}*.fit.gz",
        "-P",
        tmp_dir,
        base_url,
    ]
    subprocess.run(command, check=True)


def download_callisto_data(
    instrument_code: CallistoInstrumentCode, year: int, month: int, day: int
) -> list[str]:
    """
    Download and decompress e-Callisto FITS files, saving them as Spectre batch files.

    :param instrument_code: e-Callisto station instrument code.
    :param year: Year of the observation.
    :param month: Month of the observation.
    :param day: Day of the observation.
    :return: A list of file names of all newly created batch files, as absolute paths within the container's file system. Additionally, return the start date shared by all batch files.
    """
    tmp_dir = os.path.join(spectre_core.config.paths.get_spectre_data_dir_path(), "tmp")
    # if there are any residual files in the temporary directory, remove them.
    if os.path.exists(tmp_dir):
        shutil.rmtree(tmp_dir)
    os.makedirs(tmp_dir, exist_ok=True)

    _wget_callisto_data(instrument_code.value, year, month, day, tmp_dir)
    return sorted(_unzip_to_batches(tmp_dir))



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import spectre_core.flowgraphs

from ._validators import validate_one_of


def validate_wire_format(wire_format: str) -> None:
    validate_one_of(
        wire_format,
        [
            spectre_core.flowgraphs.USRPWireFormat.SC8,
            spectre_core.flowgraphs.USRPWireFormat.SC12,
            spectre_core.flowgraphs.USRPWireFormat.SC16,
        ],
        "wire_format",
    )


def validate_sample_rate_with_master_clock_rate(
    sample_rate: int, master_clock_rate: int
) -> None:
    """Ensure that the master clock rate is an integer multiple of the sample rate."""
    if master_clock_rate % sample_rate != 0:
        raise ValueError(
            f"The master clock rate of {master_clock_rate} [Hz] is not an integer "
            f"multiple of the sample rate {sample_rate} [Hz]."
        )



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import pydantic

import spectre_core.events
import spectre_core.flowgraphs

from ._validators import (
    skip_validator,
    validate_window_size,
    validate_nyquist_criterion,
    validate_one_of,
    validate_in_range,
)

CENTER_FREQUENCY_LOWER_BOUND = 1e6
CENTER_FREQUENCY_UPPER_BOUND = 6e9
SAMPLE_RATE_LOWER_BOUND = 1e6
SAMPLE_RATE_UPPER_BOUND = 20e6
LNA_GAIN_LOWER_BOUND = 0
LNA_GAIN_UPPER_BOUND = 40
VGA_GAIN_LOWER_BOUND = 0
VGA_GAIN_UPPER_BOUND = 62
ALLOWED_BANDWIDTHS = [
    1e6,
    2e6,
    3e6,
    4e6,
    5e6,
    6e6,
    7e6,
    8e6,
    9e6,
    10e6,
    11e6,
    12e6,
    13e6,
    14e6,
    15e6,
    16e6,
    17e6,
    18e6,
    19e6,
    20e6,
]


class HackRFOneFixedCenterFrequency(
    spectre_core.flowgraphs.HackRFFixedCenterFrequencyModel,
    spectre_core.events.FixedCenterFrequencyModel,
):
    @pydantic.model_validator(mode="after")
    def validator(self, info: pydantic.ValidationInfo):
        if skip_validator(info):
            return self
        validate_nyquist_criterion(self.sample_rate, self.bandwidth)
        validate_window_size(self.window_size)
        validate_in_range(
            self.center_frequency,
            lower_bound=CENTER_FREQUENCY_LOWER_BOUND,
            upper_bound=CENTER_FREQUENCY_UPPER_BOUND,
            name="center_frequency",
        )
        validate_in_range(
            self.sample_rate,
            lower_bound=SAMPLE_RATE_LOWER_BOUND,
            upper_bound=SAMPLE_RATE_UPPER_BOUND,
            name="center_frequency",
        )
        validate_in_range(
            self.lna_gain,
            lower_bound=LNA_GAIN_LOWER_BOUND,
            upper_bound=LNA_GAIN_UPPER_BOUND,
            name="lna_gain",
        )
        validate_in_range(
            self.vga_gain,
            lower_bound=VGA_GAIN_LOWER_BOUND,
            upper_bound=VGA_GAIN_UPPER_BOUND,
            name="vga_gain",
        )
        validate_one_of(self.bandwidth, ALLOWED_BANDWIDTHS, "bandwidth")
        return self



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import typing
import abc

import pydantic
import watchdog.events

import spectre_core.spectrograms
import spectre_core.batches
import spectre_core.fields

_LOGGER = logging.getLogger(__name__)


class BaseModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(
        validate_assignment=True,
    )
    time_range: spectre_core.fields.Field.time_range = 0
    origin: spectre_core.fields.Field.origin = "NOTSET"
    telescope: spectre_core.fields.Field.telescope = "NOTSET"
    instrument: spectre_core.fields.Field.instrument = "NOTSET"
    object: spectre_core.fields.Field.object_ = "NOTSET"
    obs_alt: spectre_core.fields.Field.obs_alt = 0.0
    obs_lat: spectre_core.fields.Field.obs_lat = 0.0
    obs_lon: spectre_core.fields.Field.obs_lon = 0.0


B = typing.TypeVar("B", bound=spectre_core.batches.Base)
M = typing.TypeVar("M", bound=BaseModel)


class Base(abc.ABC, typing.Generic[M, B], watchdog.events.FileSystemEventHandler):
    def __init__(
        self,
        tag: str,
        model: M,
        batch_cls: typing.Type[B],
        queued_file: typing.Optional[str] = None,
        cached_spectrogram: typing.Optional[
            spectre_core.spectrograms.Spectrogram
        ] = None,
    ) -> None:
        """An abstract interface enabling event-driven file processing.

        :param tag: The data tag.
        :param model: Defines configurable parameters.
        :param batch_cls: The batch used to read data files.
        :param queued_file: Optionally override the queued file, defaults to None
        :param cached_spectrogram: Optionally override the cached spectrogram, defaults to None
        """
        self._tag = tag
        self.__batch_cls = batch_cls
        self.__model = model
        self.__queued_file = queued_file
        self.__cached_spectrogram = cached_spectrogram

    @abc.abstractmethod
    def process(self, batch: B) -> spectre_core.spectrograms.Spectrogram:
        """Transform data from the input batch into a spectrogram.

        :param batch: The batch providing an interface to read signal data from the filesystem.
        :return: The signal data transformed into a spectrogram.
        """

    @property
    @abc.abstractmethod
    def _watch_extension(self) -> str:
        """Newly created files with this extension trigger the batch to be processed."""

    def on_created(self, event: watchdog.events.FileSystemEvent) -> None:
        """Process a newly created batch file, only once the next batch is created.

        Since we assume that the batches are non-overlapping in time, this guarantees
        we avoid post processing a file while it is being written to. Files are processed
        sequentially, in the order they are created.

        :param event: The file system event containing the file details.
        """
        # The `src_path`` attribute holds the absolute path of the freshly closed file
        absolute_file_path = event.src_path

        # Only process a file if:
        #
        # - It's extension matches the `watch_extension` as defined in the config.
        # - It's tag matches the current sessions tag.
        #
        # This is important for two reasons.
        #
        # In the case of one session, the capture worker may write to two batch files simultaneously
        # (e.g., raw data file + seperate metadata file). We want to process them together - but this method will get called
        # seperately for both file creation events. So, we filter by extension to account for this.
        #
        # Additionally in the case of multiple sessions, the capture workers will create batch files in the same directory concurrently.
        # This method is triggered for all file creation events, so we ensure the batch file tag matches the session tag and early return
        # otherwise. This way, each post processor worker picks up the right files to process.
        if not absolute_file_path.endswith(f"_{self._tag}.{self._watch_extension}"):
            return

        _LOGGER.info(f"Noticed {absolute_file_path}")
        # If there exists a queued file, try and process it
        if self.__queued_file is not None:
            try:
                _LOGGER.info(f"Processing {self.__queued_file}")
                batches_dir_path, start_time, tag, _ = (
                    spectre_core.batches.parse_batch_file_path(self.__queued_file)
                )
                spectrogram = self.process(
                    self.__batch_cls(batches_dir_path, start_time, tag)
                )
                self.__cache_spectrogram(spectrogram)
            except Exception:
                _LOGGER.error(
                    f"An error has occured while processing {self.__queued_file}",
                    exc_info=True,
                )
                # Flush any internally stored spectrogram on error to avoid lost data
                self.__flush_cache()
                # re-raise the exception to the main thread
                raise

        # Queue the current file for processing next
        _LOGGER.info(f"Queueing {absolute_file_path} for post processing")
        self.__queued_file = absolute_file_path

    def __cache_spectrogram(
        self, spectrogram: spectre_core.spectrograms.Spectrogram
    ) -> None:
        if self.__cached_spectrogram is None:
            self.__cached_spectrogram = spectrogram
        else:
            self.__cached_spectrogram = spectre_core.spectrograms.join_spectrograms(
                [self.__cached_spectrogram, spectrogram]
            )

        if self.__cached_spectrogram.time_range >= self.__model.time_range:
            self.__flush_cache()

    def __flush_cache(self) -> None:
        if self.__cached_spectrogram:
            _LOGGER.info(
                f"Flushing spectrogram to file with start time "
                f"'{self.__cached_spectrogram.format_start_time()}'"
            )
            self.__cached_spectrogram.save(
                self._tag,
                self.__model.origin,
                self.__model.instrument,
                self.__model.telescope,
                self.__model.object,
                self.__model.obs_alt,
                self.__model.obs_lat,
                self.__model.obs_lon,
            )
            _LOGGER.info("Flush successful, resetting spectrogram cache")
            self.__cached_spectrogram = None  # reset the cache



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
import signal
import typing
import pydantic

import gnuradio.gr


import spectre_core.config
import spectre_core.fields


class BaseModel(pydantic.BaseModel):
    """Base model to be inherited by all flowgraph models."""

    model_config = pydantic.ConfigDict(validate_assignment=True)
    max_noutput_items: spectre_core.fields.Field.max_noutput_items = 100000


M = typing.TypeVar("M", bound="BaseModel")


class Base(gnuradio.gr.top_block, typing.Generic[M]):
    def __init__(
        self,
        tag: str,
        model: M,
        batches_dir_path: typing.Optional[str] = None,
    ):
        """An abstract interface for a configurable GNU Radio flowgraph.

        :param tag: The data tag.
        :param model: The model containing configurable parameters.
        :param batches_dir_path: Directory to store data produced at runtime.
        """
        super().__init__()
        self.__model = model
        self._batches_dir_path = (
            batches_dir_path or spectre_core.config.paths.get_batches_dir_path()
        )
        self.configure(tag, self.__model)

    def configure(self, tag: str, model: M) -> None:
        """Configure the flowgraph for the block.

        :param tag: The data tag.
        :param model: Contains parameters to configure the flowgraph.
        """
        # TODO: Using the `@abc.abstractmethod` decorator causes static type checking to complain that subclasses are abstract, even
        # when they implement this method. I think inheriting from `gnuradio.gr.top_block` is throwing things off.
        raise NotImplementedError("Flowgraphs must implement the `configure` method.")

    def activate(self) -> None:
        """Activate the GNU Radio flowgraph."""

        def sig_handler(sig=None, frame=None):
            self.stop()
            self.wait()
            sys.exit(0)

        signal.signal(signal.SIGINT, sig_handler)
        signal.signal(signal.SIGTERM, sig_handler)

        self.run(self.__model.max_noutput_items)



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import typing
import datetime
import gc

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.figure
import matplotlib.axes
from matplotlib import use

import spectre_core.spectrograms
import spectre_core.config

from ._base import BasePanel, XAxisType
from ._format import PanelFormat


class PanelStack:
    """Visualise spectrogram data in a stack of panels."""

    def __init__(
        self,
        panel_format: PanelFormat = PanelFormat(),
        time_type: spectre_core.spectrograms.TimeType = spectre_core.spectrograms.TimeType.RELATIVE,
        figsize: tuple[int, int] = (15, 8),
        non_interactive: bool = False,
    ) -> None:
        """Initialize an instance of `PanelStack`.

        :param panel_format: Formatting applied across all panels in the stack. Defaults to `PanelFormat()`.
        :param time_type: The type of time assigned to spectrograms, defaults to `spectre_core.spectrograms.TimeType.RELATIVE`.
        :param figsize: The size of the `matplotlib` figure as (width, height). Defaults to (15, 8).
        """
        self._panel_format = panel_format
        self._time_type = time_type
        self._figsize = figsize

        if non_interactive:
            # Use a non-interactive matplotlib backend, which can only write files.
            use("agg")

        self._panels: list[BasePanel] = []
        self._superimposed_panels: list[BasePanel] = []

        self._fig: typing.Optional[matplotlib.figure.Figure] = None
        self._axs: typing.Optional[np.ndarray] = None

    def _sort_by_xaxis_type(self, panels: list[BasePanel]) -> list[BasePanel]:
        return list(sorted(panels, key=lambda panel: panel.xaxis_type.value))

    @property
    def panels(self) -> list[BasePanel]:
        """Get the panels in the stack, sorted by their `XAxisType`."""
        return self._sort_by_xaxis_type(self._panels)

    @property
    def superimposed_panels(self) -> list[BasePanel]:
        """Get the superimposed panels in the stack, sorted by their `XAxisType"""
        return self._sort_by_xaxis_type(self._superimposed_panels)

    @property
    def num_panels(self) -> int:
        """Get the number of panels in the stack."""
        return len(self._panels)

    @property
    def num_superimposed_panels(self) -> int:
        """Get the number of superimposed panels in the stack."""
        return len(self._superimposed_panels)

    @property
    def time_type(self) -> spectre_core.spectrograms.TimeType:
        """The imposed time type on all time series panels in the stack.

        :raises ValueError: If the `time_type` has not been set.
        """
        return self._time_type

    @time_type.setter
    def time_type(self, value: spectre_core.spectrograms.TimeType) -> None:
        """Set the `spectre_core.spectrograms.TimeType` for all time series panels in the stack.

        This controls how time is represented and annotated on the panel.

        :param value: The `spectre_core.spectrograms.TimeType` to impose on all time series panels in the stack.
        """
        self._time_type = value

    def _get_fig(self) -> matplotlib.figure.Figure:
        """Get the shared `matplotlib` figure for the panel stack.

        :raises ValueError: If the axes have not been initialized.
        """
        if self._fig is None:
            raise ValueError(
                f"An unexpected error has occured, `fig` must be set for the panel stack."
            )
        return self._fig

    def _get_axes(self) -> np.ndarray:
        """Get the `matplotlib` axes in the stack.

        :return: An array of `matplotlib.axes.Axes`, one for each panel in the stack.
        :raises ValueError: If the axes have not been initialized.
        """
        if self._axs is None:
            raise ValueError(
                f"An unexpected error has occured, `axs` must be set for the panel stack."
            )
        return np.atleast_1d(self._axs)

    def _validate_time_type(self, panel: BasePanel) -> None:
        """Check that the time type of the input panel, is consistent with the time type of the stack."""
        if panel.get_time_type() != self._time_type:
            raise ValueError(
                f"Cannot add a panel with inconsistent time type. "
                f"Expected {self._time_type.value}, but got {panel.get_time_type().value}"
            )

    def add_panel(
        self,
        panel: BasePanel,
        identifier: typing.Optional[str] = None,
        panel_format: typing.Optional[PanelFormat] = None,
    ) -> None:
        """Add a panel to the stack.

        Overrides the time type of the panel, to the time type of the stack.

        :param panel: An instance of a `BasePanel` subclass to be added to the stack.
        :param identifier: An typing.Optional string to link the panel with others for superimposing.
        """
        panel.set_panel_format(panel_format or self._panel_format)
        panel.set_time_type(self._time_type)
        if identifier:
            panel.set_identifier(identifier)

        self._panels.append(panel)

    def superimpose_panel(
        self,
        panel: BasePanel,
        identifier: typing.Optional[str] = None,
        panel_format: typing.Optional[PanelFormat] = None,
    ) -> None:
        """Superimpose a panel onto an existing panel in the stack.

        Overrides the time type of the panel, to the time type of the stack.

        :param panel: The panel to superimpose.
        :param identifier: An typing.Optional identifier to link panels during superimposing, defaults to None
        """
        if identifier:
            panel.set_identifier(identifier)
        panel.set_panel_format(panel_format or self._panel_format)
        panel.set_time_type(self._time_type)

        self._superimposed_panels.append(panel)

    def _init_plot_style(self) -> None:
        """Initialize the global plot style for the stack.

        This method sets `matplotlib` styles and font sizes based on the `panel_format`.
        """
        plt.style.use(self._panel_format.style)

        plt.rc("font", size=self._panel_format.small_size)

        plt.rc(
            "axes",
            titlesize=self._panel_format.medium_size,
            labelsize=self._panel_format.medium_size,
        )

        plt.rc("xtick", labelsize=self._panel_format.small_size)
        plt.rc("ytick", labelsize=self._panel_format.small_size)

        plt.rc("legend", fontsize=self._panel_format.small_size)
        plt.rc("figure", titlesize=self._panel_format.large_size)

    def _create_figure_and_axes(self) -> None:
        """Create the `matplotlib` figure and axes for the panel stack.

        This initializes a figure with a specified number of vertically stacked axes.
        """
        self._fig, self._axs = plt.subplots(
            self.num_panels, 1, figsize=self._figsize, layout="constrained"
        )

    def _assign_axes(self) -> None:
        """Assign each axes in the figure to some panel in the stack.

        Axes are shared between panels with common `XAxisType`.
        """
        shared_axes: dict[XAxisType, matplotlib.axes.Axes] = {}
        for i, panel in enumerate(self.panels):
            ax = self._get_axes()[i]
            panel.set_ax(ax)
            panel.set_fig(self._fig)
            if panel.xaxis_type in shared_axes:
                panel.sharex(shared_axes[panel.xaxis_type])
            else:
                shared_axes[panel.xaxis_type] = ax

    def _overlay_superimposed_panels(self) -> None:
        """Superimpose panels onto matching panels in the stack.

        For each superimposed panel, find a matching panel in the stack with the same name
        and identifier. Share the axes and figure, then draw the superimposed panel. If the
        panel contains cuts, overlay those cuts onto the corresponding spectrogram panels.
        """
        for super_panel in self._superimposed_panels:
            for panel in self._panels:
                if panel.name == super_panel.name and (
                    panel.get_identifier() == super_panel.get_identifier()
                ):
                    super_panel.share_axes(panel)
                    super_panel.draw()

    def _make_figure(self) -> None:
        """Make the panel stack figure."""
        if self.num_panels < 1:
            raise ValueError(f"There must be at least one panel in the stack.")

        self._init_plot_style()
        self._create_figure_and_axes()
        self._assign_axes()

        last_panel_per_axis = {panel.xaxis_type: panel for panel in self.panels}

        for panel in self.panels:
            panel.draw()
            panel.annotate_yaxis()

            if panel == last_panel_per_axis[panel.xaxis_type]:
                panel.annotate_xaxis()
            else:
                panel.hide_xaxis_labels()

        self._overlay_superimposed_panels()

    def _close(self) -> None:
        """Prevent memory leaks once a figure has been created, and successfully visualised."""
        self._get_fig().clear()
        plt.close(self._fig)
        # Garbage collection seems to be required to prevent the memory leak.
        # See https://github.com/jcfitzpatrick12/spectre/issues/128
        gc.collect()

    def show(self) -> None:
        """Display the panel stack figure."""
        self._make_figure()
        self._get_fig().show()
        self._close()

    def save(self, tag: str, batches_dir_path: typing.Optional[str] = None) -> str:
        """Save the panel stack figure as a batch file under the input tag.

        :return: The file path of the newly created batch file containing the figure.
        """
        self._make_figure()
        first_panel = self._panels[0]

        start_dt = typing.cast(
            datetime.datetime,
            first_panel.spectrogram.start_datetime.astype(datetime.datetime),
        )
        batch_name = (
            f"{start_dt.strftime(spectre_core.config.TimeFormat.DATETIME)}_{tag}"
        )
        batch_file_path = os.path.join(
            batches_dir_path
            or spectre_core.config.paths.get_batches_dir_path(
                start_dt.year, start_dt.month, start_dt.day
            ),
            f"{batch_name}.png",
        )
        # If the parent directory does not exist, create it.
        os.makedirs(os.path.dirname(batch_file_path), exist_ok=True)
        self._get_fig().savefig(batch_file_path)
        self._close()
        return batch_file_path



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import dataclasses
import os
import typing
import json
import enum

import spectre_core.io
import spectre_core.config


class _ReservedTagStr(enum.Enum):
    """Reserve substrings to indicate batch files derived from external spectrogram data."""

    CALLISTO = "callisto"


def _validate_tag(tag: str) -> None:
    if "_" in tag:
        raise ValueError("An underscore is not allowed in a config tag.")
    for s in _ReservedTagStr:
        if s.value in tag:
            raise ValueError(
                f"'{tag}' is an invalid tag, the substring '{s}' is reserved."
            )


def _validate_keys(content: dict[str, str]) -> None:
    required_keys = {
        _CaptureConfigKey.RECEIVER_NAME,
        _CaptureConfigKey.RECEIVER_MODE,
        _CaptureConfigKey.PARAMETERS,
    }
    missing_keys = required_keys - content.keys()
    if missing_keys:
        raise ValueError(f"Missing required keys in config content: {missing_keys}")


@dataclasses.dataclass(frozen=True)
class _CaptureConfigKey:
    """Defined JSON keys for configs.

    :ivar RECEIVER_NAME: The name of the receiver.
    :ivar RECEIVER_MODE: The operating mode for the receiver.
    :ivar PARAMETERS: User-configured parameters.
    """

    RECEIVER_NAME = "receiver_name"
    RECEIVER_MODE = "receiver_mode"
    PARAMETERS = "parameters"


class Config:
    def __init__(self, tag: str, content: dict[str, typing.Any]) -> None:
        """A container for config data.

        :param tag: The tag of the config.
        :param content: The config data.
        """
        _validate_keys(content)
        _validate_tag(tag)
        self._tag = tag
        self._content = content

    @property
    def tag(self) -> str:
        """A config tag."""
        return self._tag

    @property
    def receiver_name(self) -> str:
        """The name of the receiver."""
        return self._content[_CaptureConfigKey.RECEIVER_NAME]

    @property
    def receiver_mode(self) -> str:
        """The operating mode for the receiver."""
        return self._content[_CaptureConfigKey.RECEIVER_MODE]

    @property
    def parameters(self) -> dict[str, typing.Any]:
        """Configurable parameters."""
        return self._content[_CaptureConfigKey.PARAMETERS]

    @property
    def content(self) -> dict[str, typing.Any]:
        return self._content


def get_config_file_path(
    tag: str, configs_dir_path: typing.Optional[str] = None
) -> str:
    """Build a config file path.

    :param tag: The config tag.
    :param configs_dir_path: Optionally override the directory containing the configs, defaults to None
    """
    configs_dir_path = (
        configs_dir_path or spectre_core.config.paths.get_configs_dir_path()
    )
    return os.path.join(configs_dir_path, f"{tag}.json")


def read_config(tag: str, configs_dir_path: typing.Optional[str] = None) -> Config:
    """Read any config data from the filesystem for any receiver, without validation.

    :param tag: The config tag.
    :param configs_dir_path: Optionally override the directory containing the configs, defaults to None
    :return: A container storing the config data.
    """
    config_file_path = get_config_file_path(tag, configs_dir_path)
    if not os.path.exists(config_file_path):
        raise FileNotFoundError(f"A config with tag '{tag}' does not exist.")
    return Config(
        tag,
        spectre_core.io.read_file(config_file_path, spectre_core.io.FileFormat.JSON),
    )


def parse_config_file_name(file_name: str) -> tuple[str, str]:
    """Parse a config file name into its tag and extension."""
    return os.path.splitext(file_name)


def write_config(
    tag: str,
    receiver_name: str,
    receiver_mode: str,
    parameters: dict[str, typing.Any],
    configs_dir_path: typing.Optional[str] = None,
) -> None:
    """Write parameters to a config on the filesystem.

    :param tag: The config tag.
    :param receiver_name: The name of the receiver.
    :param receiver_mode: The name of the operating mode.
    :param parameters: The parameters to save.
    :param configs_dir_path: Optionally override the directory containing the configs, defaults to None
    """
    _validate_tag(tag)
    configs_dir_path = (
        configs_dir_path or spectre_core.config.paths.get_configs_dir_path()
    )
    content = {
        _CaptureConfigKey.RECEIVER_NAME: receiver_name,
        _CaptureConfigKey.RECEIVER_MODE: receiver_mode,
        _CaptureConfigKey.PARAMETERS: parameters,
    }
    file_path = os.path.join(configs_dir_path, f"{tag}.json")
    with open(file_path, "w") as f:
        json.dump(content, f, indent=4)



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from ._register import register_receiver
from ._base import Base
from ._names import ReceiverName


@register_receiver(ReceiverName.CUSTOM)
class Custom(Base):
    """A customisable receiver with no pre-defined operating modes.

    Use `add_mode` to add new operating modes.
    """



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import json
import abc
import base64
import typing
import enum

from typing import Any, Optional, TypeVar, Generic

T = TypeVar("T")


class FileFormat(enum.Enum):
    JSON = "json"
    TEXT = "txt"
    PNG = "png"


class Base(abc.ABC, Generic[T]):
    def __init__(self, file_path: str) -> None:
        """An abstract interface to read a file of arbitrary format from the file system.

        :param file_path: The path to the file on the filesystem.
        """
        self._cache: Optional[T] = None
        self._file_path = file_path

    @abc.abstractmethod
    def read(self) -> T:
        """Read the contents of the file."""

    @property
    def file_path(self) -> str:
        """The path to the file on the filesystem."""
        return self._file_path

    @property
    def parent_dir_path(self) -> str:
        """Return the parent directory path for the file."""
        return os.path.dirname(self._file_path)

    @property
    def base_file_name(self) -> str:
        """Return the file name without the extension."""
        return os.path.splitext(os.path.basename(self._file_path))[0]

    @property
    def extension(self) -> Optional[str]:
        """Return the file extension (without the dot), or None if no extension exists."""
        _, ext = os.path.splitext(self._file_path)
        return ext.strip(".") if ext else None

    @property
    def file_name(self) -> str:
        """Return the full file name (base file name and extension)."""
        return os.path.basename(self._file_path)

    @property
    def exists(self) -> bool:
        """Check if the file exists in the filesystem."""
        return os.path.exists(self._file_path)

    def cached_read(self) -> T:
        """Read the file contents, caching the result for repeated calls.

        :return: The cached contents of the file.
        """
        if self._cache is None:
            self._cache = self.read()
        return self._cache

    def delete(self, ignore_if_missing: bool = False) -> None:
        """Delete the file from the filesystem.

        :param ignore_if_missing: If True, skips deletion if the file does not exist, defaults to False.
        :raises FileNotFoundError: If the file is missing and `ignore_if_missing` is False.
        """
        if not self.exists:
            if not ignore_if_missing:
                raise FileNotFoundError(f"{self.file_name} does not exist.")
        else:
            os.remove(self._file_path)


class _JsonFile(Base[dict[str, Any]]):
    def read(self) -> dict[str, Any]:
        with open(self._file_path, "r", encoding="utf-8") as f:
            return json.load(f)


class _TextFile(Base[str]):
    def read(self) -> str:
        with open(self._file_path, "r", encoding="utf-8") as f:
            return f.read()


class _PngFile(Base[str]):
    def read(self) -> str:
        with open(self.file_path, "rb") as f:
            encoded = base64.b64encode(f.read())
            return encoded.decode("ascii")


@typing.overload
def read_file(
    file_path: str, file_format: typing.Literal[FileFormat.JSON]
) -> dict[str, Any]: ...
@typing.overload
def read_file(file_path: str, file_format: typing.Literal[FileFormat.TEXT]) -> str: ...
@typing.overload
def read_file(file_path: str, file_format: typing.Literal[FileFormat.PNG]) -> str: ...


def read_file(file_path: str, file_format: FileFormat) -> typing.Any:
    """Read a file using the specified file format.

    :param file_path: The path to the file.
    :param file_format: The format of the file (FileFormat enum).
    :return: The contents of the file.
    :raises ValueError: If the file format is unsupported.
    """
    file: Base
    if file_format == FileFormat.JSON:
        file = _JsonFile(file_path)
    elif file_format == FileFormat.TEXT:
        file = _TextFile(file_path)
    elif file_format == FileFormat.PNG:
        file = _PngFile(file_path)
    else:
        raise ValueError(f"Unsupported file format: {file_format}")
    return file.read()



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""Configurable, extensible GNURadio flowgraphs."""


from ._base import Base
from ._signal_generator import (
    SignalGeneratorCosineWave,
    SignalGeneratorCosineWaveModel,
    SignalGeneratorConstantStaircase,
    SignalGeneratorConstantStaircaseModel,
)
from ._rsp1a import (
    RSP1AFixedCenterFrequency,
    RSP1AFixedCenterFrequencyModel,
    RSP1ASweptCenterFrequency,
    RSP1ASweptCenterFrequencyModel,
)
from ._rspduo import (
    RSPduoFixedCenterFrequency,
    RSPduoFixedCenterFrequencyModel,
    RSPduoSweptCenterFrequency,
    RSPduoSweptCenterFrequencyModel,
    RSPduoPort,
)
from ._rspdx import (
    RSPdxFixedCenterFrequency,
    RSPdxFixedCenterFrequencyModel,
    RSPdxSweptCenterFrequency,
    RSPdxSweptCenterFrequencyModel,
    RSPdxPort,
)
from ._usrp import (
    USRPFixedCenterFrequency,
    USRPFixedCenterFrequencyModel,
    USRPSweptCenterFrequency,
    USRPSweptCenterFrequencyModel,
    USRPWireFormat,
)
from ._hackrf import HackRFFixedCenterFrequency, HackRFFixedCenterFrequencyModel
from ._rtlsdr import RTLSDRFixedCenterFrequency, RTLSDRFixedCenterFrequencyModel

__all__ = [
    "Base",
    "SignalGeneratorCosineWave",
    "SignalGeneratorCosineWaveModel",
    "SignalGeneratorConstantStaircase",
    "SignalGeneratorConstantStaircaseModel",
    "RSP1AFixedCenterFrequency",
    "RSP1AFixedCenterFrequencyModel",
    "RSP1ASweptCenterFrequency",
    "RSP1ASweptCenterFrequencyModel",
    "RSPduoFixedCenterFrequency",
    "RSPduoFixedCenterFrequencyModel",
    "RSPduoSweptCenterFrequency",
    "RSPduoSweptCenterFrequencyModel",
    "RSPduoPort",
    "RSPdxFixedCenterFrequency",
    "RSPdxFixedCenterFrequencyModel",
    "RSPdxSweptCenterFrequency",
    "RSPdxSweptCenterFrequencyModel",
    "RSPdxPort",
    "USRPFixedCenterFrequency",
    "USRPFixedCenterFrequencyModel",
    "USRPSweptCenterFrequency",
    "USRPSweptCenterFrequencyModel",
    "USRPWireFormat",
    "HackRFFixedCenterFrequency",
    "HackRFFixedCenterFrequencyModel",
    "RTLSDRFixedCenterFrequency",
    "RTLSDRFixedCenterFrequencyModel",
]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import typing
import collections

import spectre_core.io
import spectre_core.config


def parse_log_file_name(file_name: str) -> tuple[str, str, str]:
    """Parse the file name of a log into a start time, process ID and process type."""
    base_file_name, _ = os.path.splitext(file_name)
    log_start_time, pid, process_type = base_file_name.split("_")
    return log_start_time, pid, process_type


class Log(spectre_core.io.Base):
    def __init__(self, file_path: str) -> None:
        """Interface to read log files generated by Spectre.

        :param file_path: The file path of the log.
        """
        super().__init__(file_path)
        self._start_time, self._pid, self._process_type = parse_log_file_name(
            self.file_name
        )

    @property
    def start_time(self) -> str:
        """The system time when the log was created."""
        return self._start_time

    @property
    def pid(self) -> str:
        """The ID of the process writing to the log file."""
        return self._pid

    @property
    def process_type(self) -> str:
        """Indicates the type of process."""
        return self._process_type

    def read(self) -> str:
        """Read the contents of the log file."""
        return spectre_core.io.read_file(
            self.file_path, spectre_core.io.FileFormat.TEXT
        )


class Logs:
    def __init__(
        self,
        process_type: typing.Optional[str] = None,
        logs_dir_path: typing.Optional[str] = None,
    ) -> None:
        """An interface to sort and read logs generated by Spectre.

        :param process_type: Optionally filter by process type, defaults to None
        :param logs_dir_path: Optionally override the directory containing the logs, defaults to None
        """

        self.process_type = process_type
        self._logs_dir_path = (
            logs_dir_path or spectre_core.config.paths.get_logs_dir_path()
        )

        self._log_map: dict[str, Log] = collections.OrderedDict()
        self.__update()

    def __update(self) -> None:
        """Perform a fresh search of all files according to the date and process type."""
        paths = []
        for root, _, files in os.walk(self._logs_dir_path):
            for file in files:
                paths.append(os.path.join(root, file))

        for path in paths:
            file_name = os.path.basename(path)
            _, _, process_type = parse_log_file_name(file_name)

            if self.process_type and process_type != self.process_type:
                continue

            self._log_map[file_name] = Log(path)

        self._log_map = collections.OrderedDict(sorted(self._log_map.items()))

    def __iter__(self) -> typing.Iterator[Log]:
        yield from self._log_map.values()

    def get_from_file_name(self, file_name: str) -> Log:
        """Retrieve a log based on the file name.

        :param file_name: The name of the log file (with or without extension).
        :raises FileNotFoundError: If the log file name is not found.
        :return: The log with the file name.
        """
        # auto strip the extension if present
        file_name, _ = os.path.splitext(file_name)
        try:
            return self._log_map[file_name]
        except KeyError:
            raise FileNotFoundError(f"Log not found with file name '{file_name}'")

    def get_from_pid(self, pid: str) -> Log:
        """Retrieve a log based on the process ID.

        :param pid: The process ID to search for.
        :raises FileNotFoundError: If a log file corresponding to the process ID is not found.
        :return: The log corresponding to the process ID.
        """
        for log in self:
            if log.pid == pid:
                return log
        raise FileNotFoundError(f"Log not found with pid '{pid}'")



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""Download third-party spectrogram data."""

from ._callisto import download_callisto_data, CallistoInstrumentCode

__all__ = ["download_callisto_data", "CallistoInstrumentCode"]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""Shared pydantic fields for all models."""

from ._fields import Field

__all__ = ["Field"]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import typing

import spectre_core.exceptions
import spectre_core.batches

from ._register import receivers
from ._base import Base
from ._config import _ReservedTagStr, read_config
from ._signal_generator import SignalGenerator
from ._custom import Custom
from ._rsp1a import RSP1A
from ._rspduo import RSPduo
from ._rspdx import RSPdx
from ._usrp import USRP
from ._b200mini import B200mini
from ._hackrf import HackRF
from ._hackrfone import HackRFOne
from ._rtlsdr import RTLSDR


@typing.overload
def get_receiver(
    receiver_name: typing.Literal["custom"], mode: typing.Optional[str] = None
) -> Custom: ...
@typing.overload
def get_receiver(
    receiver_name: typing.Literal["signal_generator"], mode: typing.Optional[str] = None
) -> SignalGenerator: ...
@typing.overload
def get_receiver(
    receiver_name: typing.Literal["rsp1a"], mode: typing.Optional[str] = None
) -> RSP1A: ...
@typing.overload
def get_receiver(
    receiver_name: typing.Literal["rspduo"], mode: typing.Optional[str] = None
) -> RSPduo: ...
@typing.overload
def get_receiver(
    receiver_name: typing.Literal["rspdx"], mode: typing.Optional[str] = None
) -> RSPdx: ...
@typing.overload
def get_receiver(
    receiver_name: typing.Literal["usrp"], mode: typing.Optional[str] = None
) -> USRP: ...
@typing.overload
def get_receiver(
    receiver_name: typing.Literal["b200mini"], mode: typing.Optional[str] = None
) -> B200mini: ...
@typing.overload
def get_receiver(
    receiver_name: typing.Literal["hackrf"], mode: typing.Optional[str] = None
) -> HackRF: ...
@typing.overload
def get_receiver(
    receiver_name: typing.Literal["hackrfone"], mode: typing.Optional[str] = None
) -> HackRFOne: ...
@typing.overload
def get_receiver(
    receiver_name: typing.Literal["rtlsdr"], mode: typing.Optional[str] = None
) -> RTLSDR: ...
@typing.overload
def get_receiver(receiver_name: str, mode: typing.Optional[str] = None) -> Base: ...


def get_receiver(receiver_name: str, mode: typing.Optional[str] = None) -> Base:
    """Get a registered receiver.

    :param receiver_name: The name of the receiver.
    :param mode: The initial operating mode for the receiver, defaults to None
    :raises ReceiverNotFoundError: If the receiver name is not registered.
    :return: An instance of the receiver class registered under `receiver_name`.
    """
    receiver_cls = receivers.get(receiver_name)
    if receiver_cls is None:
        valid_receivers = list(receivers.keys())
        raise spectre_core.exceptions.ReceiverNotFoundError(
            f"Could not find the receiver '{receiver_name}', "
            f"expected one of {valid_receivers}"
        )
    return receiver_cls(receiver_name, mode=mode)


def get_batch_cls(
    tag: str, configs_dir_path: typing.Optional[str] = None
) -> typing.Type[spectre_core.batches.Base]:
    """Get the right API used to read batch files under the input tag, accounting for
    batch files containing third-party spectrogram data.

    :param tag: The batch file tag.
    """
    for s in _ReservedTagStr:
        if s.value in tag and s == _ReservedTagStr.CALLISTO:
            return spectre_core.batches.CallistoBatch

    config = read_config(tag, configs_dir_path)
    receiver = get_receiver(config.receiver_name, config.receiver_mode)
    return receiver.batch_cls



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import time
import dataclasses

from gnuradio import spectre
from gnuradio import uhd

import spectre_core.fields

from ._base import Base, BaseModel


@dataclasses.dataclass(frozen=True)
class USRPWireFormat:
    """Indicates the form of the data over the bus/network."""

    SC8 = "sc8"
    SC12 = "sc12"
    SC16 = "sc16"


class USRPFixedCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 600000
    batch_size: spectre_core.fields.Field.batch_size = 3
    center_frequency: spectre_core.fields.Field.center_frequency = 95.8e6
    bandwidth: spectre_core.fields.Field.bandwidth = 600000
    gain: spectre_core.fields.Field.gain = 35
    wire_format: spectre_core.fields.Field.wire_format = USRPWireFormat.SC12
    master_clock_rate: spectre_core.fields.Field.master_clock_rate = 60000000


class USRPFixedCenterFrequency(Base[USRPFixedCenterFrequencyModel]):
    def configure(self, tag: str, model: USRPFixedCenterFrequencyModel) -> None:
        master_clock_rate = f"master_clock_rate={model.master_clock_rate}"
        self.uhd_usrp_source = uhd.usrp_source(
            ",".join(("", "", master_clock_rate)),
            uhd.stream_args(
                cpu_format="fc32",
                otw_format=model.wire_format,
                args="",
                channels=[0],
            ),
        )
        self.uhd_usrp_source.set_samp_rate(model.sample_rate)
        self.uhd_usrp_source.set_time_now(uhd.time_spec(time.time()), uhd.ALL_MBOARDS)
        self.uhd_usrp_source.set_center_freq(model.center_frequency, 0)
        self.uhd_usrp_source.set_bandwidth(model.bandwidth, 0)
        self.uhd_usrp_source.set_rx_agc(False, 0)
        self.uhd_usrp_source.set_auto_dc_offset(False, 0)
        self.uhd_usrp_source.set_auto_iq_balance(False, 0)
        self.uhd_usrp_source.set_gain(model.gain, 0)
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.batch_size,
            model.sample_rate,
            False,
            "rx_freq",
            0,
        )
        self.connect((self.uhd_usrp_source, 0), (self.spectre_batched_file_sink, 0))


class USRPSweptCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 2000000
    batch_size: spectre_core.fields.Field.batch_size = 3
    bandwidth: spectre_core.fields.Field.bandwidth = 2e6
    min_frequency: spectre_core.fields.Field.min_frequency = 95e6
    max_frequency: spectre_core.fields.Field.max_frequency = 101e6
    samples_per_step: spectre_core.fields.Field.samples_per_step = 60000
    frequency_step: spectre_core.fields.Field.frequency_step = 2e6
    gain: spectre_core.fields.Field.gain = 35
    wire_format: spectre_core.fields.Field.wire_format = USRPWireFormat.SC12
    master_clock_rate: spectre_core.fields.Field.master_clock_rate = 60000000


class USRPSweptCenterFrequency(Base[USRPSweptCenterFrequencyModel]):
    def configure(self, tag: str, model: USRPSweptCenterFrequencyModel) -> None:
        master_clock_rate = f"master_clock_rate={model.master_clock_rate}"
        self.uhd_usrp_source = uhd.usrp_source(
            ",".join(("", "", master_clock_rate)),
            uhd.stream_args(
                cpu_format="fc32",
                otw_format=model.wire_format,
                args="",
                channels=[0],
            ),
        )
        self.uhd_usrp_source.set_samp_rate(model.sample_rate)
        self.uhd_usrp_source.set_time_now(uhd.time_spec(time.time()), uhd.ALL_MBOARDS)
        self.uhd_usrp_source.set_center_freq(model.min_frequency, 0)
        self.uhd_usrp_source.set_bandwidth(model.bandwidth, 0)
        self.uhd_usrp_source.set_rx_agc(False, 0)
        self.uhd_usrp_source.set_auto_dc_offset(False, 0)
        self.uhd_usrp_source.set_auto_iq_balance(False, 0)
        self.uhd_usrp_source.set_gain(model.gain, 0)

        self.spectre_sweep_driver = spectre.sweep_driver(
            model.min_frequency,
            model.max_frequency,
            model.frequency_step,
            model.sample_rate,
            model.samples_per_step,
            "freq",
        )

        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.batch_size,
            model.sample_rate,
            True,
            "rx_freq",
            model.min_frequency,
        )

        self.msg_connect(
            (self.spectre_sweep_driver, "retune_command"),
            (self.uhd_usrp_source, "command"),
        )
        self.connect((self.uhd_usrp_source, 0), (self.spectre_batched_file_sink, 0))
        self.connect((self.uhd_usrp_source, 0), (self.spectre_sweep_driver, 0))



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later


"""Templatise configurable parameters."""

from ._signal_generator import (
    SignalGeneratorCosineWaveModel,
    SignalGeneratorConstantStaircaseModel,
)
from ._rsp1a import RSP1AFixedCenterFrequency, RSP1ASweptCenterFrequency
from ._rspduo import RSPduoFixedCenterFrequency, RSPduoSweptCenterFrequency
from ._rspdx import RSPdxFixedCenterFrequency, RSPdxSweptCenterFrequency
from ._usrp import USRPFixedCenterFrequency, USRPSweptCenterFrequency
from ._b200mini import B200miniFixedCenterFrequency, B200miniSweptCenterFrequency
from ._hackrf import HackRFFixedCenterFrequency
from ._hackrfone import HackRFOneFixedCenterFrequency
from ._rtlsdr import RTLSDRFixedCenterFrequency

__all__ = [
    "SignalGeneratorCosineWaveModel",
    "SignalGeneratorConstantStaircaseModel",
    "RSP1AFixedCenterFrequency",
    "RSP1ASweptCenterFrequency",
    "RSPduoFixedCenterFrequency",
    "RSPduoSweptCenterFrequency",
    "RSPdxFixedCenterFrequency",
    "RSPdxSweptCenterFrequency",
    "USRPFixedCenterFrequency",
    "USRPSweptCenterFrequency",
    "B200miniFixedCenterFrequency",
    "B200miniSweptCenterFrequency",
    "HackRFFixedCenterFrequency",
    "HackRFOneFixedCenterFrequency",
    "RTLSDRFixedCenterFrequency",
]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""An API for plotting spectrogram data."""

from ._format import PanelFormat
from ._base import BasePanel, BaseTimeSeriesPanel, XAxisType
from ._panels import (
    SpectrogramPanel,
    FrequencyCutsPanel,
    TimeCutsPanel,
    IntegralOverFrequencyPanel,
)
from ._panel_names import PanelName
from ._panel_stack import PanelStack

__all__ = [
    "BaseTimeSeriesPanel",
    "XAxisType",
    "BasePanel",
    "PanelFormat",
    "PanelStack",
    "PanelName",
    "SpectrogramPanel",
    "FrequencyCutsPanel",
    "TimeCutsPanel",
    "IntegralOverFrequencyPanel",
]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import pydantic

import spectre_core.events
import spectre_core.flowgraphs

from ._validators import (
    skip_validator,
    validate_window_size,
    validate_nyquist_criterion,
)


class HackRFFixedCenterFrequency(
    spectre_core.flowgraphs.HackRFFixedCenterFrequencyModel,
    spectre_core.events.FixedCenterFrequencyModel,
):
    @pydantic.model_validator(mode="after")
    def validator(self, info: pydantic.ValidationInfo):
        if skip_validator(info):
            return self
        validate_nyquist_criterion(self.sample_rate, self.bandwidth)
        validate_window_size(self.window_size)
        return self



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from gnuradio import gr
from gnuradio import blocks
from gnuradio import analog
from gnuradio import spectre

import spectre_core.fields

from ._base import Base, BaseModel


class SignalGeneratorCosineWaveModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 128000
    batch_size: spectre_core.fields.Field.batch_size = 3
    frequency: spectre_core.fields.Field.frequency = 32000
    amplitude: spectre_core.fields.Field.amplitude = 1


class SignalGeneratorCosineWave(Base[SignalGeneratorCosineWaveModel]):
    def configure(self, tag: str, model: SignalGeneratorCosineWaveModel) -> None:
        """Record a complex-valued cosine signal to batched data files."""
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path, tag, model.batch_size, model.sample_rate
        )
        self.blocks_throttle_0 = blocks.throttle(
            gr.sizeof_float * 1, model.sample_rate, True
        )
        self.blocks_throttle_1 = blocks.throttle(
            gr.sizeof_float * 1, model.sample_rate, True
        )
        self.blocks_null_source = blocks.null_source(gr.sizeof_float * 1)
        self.blocks_float_to_complex = blocks.float_to_complex(1)
        self.analog_sig_source = analog.sig_source_f(
            model.sample_rate,
            analog.GR_COS_WAVE,
            model.frequency,
            model.amplitude,
            0,
            0,
        )

        self.connect((self.analog_sig_source, 0), (self.blocks_throttle_0, 0))
        self.connect((self.blocks_null_source, 0), (self.blocks_throttle_1, 0))
        self.connect((self.blocks_throttle_0, 0), (self.blocks_float_to_complex, 0))
        self.connect((self.blocks_throttle_1, 0), (self.blocks_float_to_complex, 1))
        self.connect(
            (self.blocks_float_to_complex, 0), (self.spectre_batched_file_sink, 0)
        )


class SignalGeneratorConstantStaircaseModel(BaseModel):
    step_increment: spectre_core.fields.Field.step_increment = 200
    sample_rate: spectre_core.fields.Field.sample_rate = 128000
    min_samples_per_step: spectre_core.fields.Field.min_samples_per_step = 4000
    max_samples_per_step: spectre_core.fields.Field.max_samples_per_step = 5000
    frequency_step: spectre_core.fields.Field.frequency_step = 128000
    batch_size: spectre_core.fields.Field.batch_size = 3


class SignalGeneratorConstantStaircase(Base[SignalGeneratorConstantStaircaseModel]):
    def configure(self, tag: str, model: SignalGeneratorConstantStaircaseModel) -> None:
        """Record a constant signal that periodically increments in value.
        Each step increases in duration, up to a maximum, before resetting."""
        self.spectre_constant_staircase = spectre.tagged_staircase(
            model.min_samples_per_step,
            model.max_samples_per_step,
            model.frequency_step,
            model.step_increment,
            model.sample_rate,
        )
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.batch_size,
            model.sample_rate,
            True,
            "rx_freq",
            0,
        )  # zero means the center frequency is unset
        self.blocks_throttle = blocks.throttle(
            gr.sizeof_gr_complex * 1, model.sample_rate, True
        )

        self.connect((self.spectre_constant_staircase, 0), (self.blocks_throttle, 0))
        self.connect((self.blocks_throttle, 0), (self.spectre_batched_file_sink, 0))



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import time
import typing
import multiprocessing

import spectre_core.logs
import spectre_core.config
from ._duration import Duration

_LOGGER = logging.getLogger(__name__)


def _make_daemon_process(
    name: str, target: typing.Callable[[], None]
) -> multiprocessing.Process:
    """
    Creates and returns a daemon `multiprocessing.Process` instance.

    :param name: The name to assign to the process.
    :param target: The function to execute in the process.
    :return: A `multiprocessing.Process` instance configured as a daemon.
    """
    return multiprocessing.Process(target=target, name=name, daemon=True)


class Worker:
    def __init__(self, name: str, target: typing.Callable[[], None]) -> None:
        """A lightweight wrapper for a `multiprocessing.Process` daemon.

        Provides a very simple API to start, kill, and restart a multiprocessing process.

        :param name: The name assigned to the process.
        :param target: The callable to be executed by the worker process.
        """
        self._name = name
        self._target = target
        self._process = _make_daemon_process(name, target)

    @property
    def name(self) -> str:
        """Get the name of the worker process.

        :return: The name of the multiprocessing process.
        """
        return self._process.name

    @property
    def is_alive(self) -> bool:
        """Return whether the managed process is alive."""
        return self._process.is_alive()

    def start(self) -> None:
        """Start the worker process.

        This method runs the `target` in the background as a daemon.
        """
        if self.is_alive:
            raise RuntimeError("A worker cannot be started twice.")

        self._process.start()

    def kill(self) -> None:
        """Kill the managed process."""
        if not self.is_alive:
            raise RuntimeError("Cannot kill a process which is not alive.")

        self._process.kill()

    def restart(self) -> None:
        """Restart the worker process.

        Kills the existing process if it is alive and then starts a new process
        after a brief pause.
        """
        _LOGGER.info(f"Restarting {self.name} worker")
        if self.is_alive:
            # forcibly stop if it is still alive
            self.kill()

        # a moment of respite
        time.sleep(0.5 * Duration.ONE_SECOND)

        # make a new process, as we can't start the same process again.
        self._process = _make_daemon_process(self._name, self._target)
        self.start()


# TODO: Somehow statically type check that `args` match the arguments to `target`
def make_worker(
    name: str,
    target: typing.Callable[..., None],
    args: tuple = (),
    configure_logging: bool = True,
    spectre_data_dir_path: typing.Optional[str] = None,
) -> Worker:
    """Create a `Worker` instance to manage a target function in a multiprocessing background daemon process.

    This function returns a `Worker` that is configured to run the given target function with the provided arguments
    in a separate process. The worker is not started automatically; you must call `start()` to call the target. The target should not return anything,
    as its return value will be discarded.

    :param name: Human-readable name for the worker process.
    :param target: The function to be executed by the worker process.
    :param args: Arguments to pass to the target function.
    :param configure_logging: If True, configure the root logger to write log events to file. Defaults to True.
    :param spectre_data_dir_path: If specified, override the `SPECTRE_DATA_DIR_PATH` environment variable to this value in the process
    managed by the worker.
    :return: A `Worker` instance managing the background process (not started).
    """

    def _worker_target() -> None:
        if spectre_data_dir_path is not None:
            spectre_core.config.paths.set_spectre_data_dir_path(spectre_data_dir_path)

        if configure_logging:
            spectre_core.logs.configure_root_logger(
                spectre_core.logs.ProcessType.WORKER,
            )

        target(*args)

    return Worker(name, _worker_target)



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later


"""Program-wide config."""

from ._paths import paths, Paths
from ._time_formats import TimeFormat

__all__ = [
    "paths",
    "Paths",
    "TimeFormat",
]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import logging
import datetime
import typing

import spectre_core.config
from ._process_types import ProcessType


def configure_root_logger(
    process_type: ProcessType,
    level: int = logging.INFO,
    logs_dir_path: typing.Optional[str] = None,
) -> str:
    """Configure the root logger to write log events to a file.

    The log file will be created at the following path:

        `<logs_dir_path>/<start_time>_<pid>_<process_type>.log`

    :param process_type: The type of the process (e.g., USER, WORKER).
    :param level: The logging level, defaults to logging.INFO.
    :param logs_dir_path: Optionally override the directory in which logs will be written, defaults to None.
    :return: The file path of the log file on the filesystem.
    """
    # Get the root logger, set its level and remove any existing handlers.
    logger = logging.getLogger()
    logger.setLevel(level)
    for handler in logger.handlers:
        logger.removeHandler(handler)

    # Get the current system time and current process ID.
    system_datetime = datetime.datetime.now()
    start_time = system_datetime.strftime(spectre_core.config.TimeFormat.DATETIME)
    pid = str(os.getpid())

    # Make the file path.
    if logs_dir_path is None:
        logs_dir_path = spectre_core.config.paths.get_logs_dir_path(
            system_datetime.year, system_datetime.month, system_datetime.day
        )
    if not os.path.exists(logs_dir_path):
        os.makedirs(logs_dir_path)
    file_path = os.path.join(
        logs_dir_path, f"{start_time}_{pid}_{process_type.value}.log"
    )

    # Add the file handler to the root logger.
    file_handler = logging.FileHandler(file_path)
    file_handler.setLevel(level)
    formatter = logging.Formatter(
        "[%(asctime)s] [%(levelname)8s] --- %(message)s (%(name)s:%(lineno)s)"
    )
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    return os.path.abspath(file_path)


def get_root_logger_state() -> tuple[bool, int]:
    """Get the state of the root logger.

    :return: A tuple containing:
        - A boolean indicating whether the root logger has any handlers.
        - The logging level of the root logger.
    """
    root_logger = logging.getLogger()
    if root_logger.handlers:
        return True, root_logger.level
    return False, logging.NOTSET



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""Create and transform spectrogram data."""

from ._spectrogram import Spectrogram, FrequencyCut, TimeCut, SpectrumUnit, TimeType
from ._transform import (
    frequency_chop,
    time_chop,
    frequency_average,
    time_average,
    join_spectrograms,
)

__all__ = [
    "Spectrogram",
    "FrequencyCut",
    "TimeCut",
    "SpectrumUnit",
    "frequency_chop",
    "time_chop",
    "frequency_average",
    "time_average",
    "join_spectrograms",
    "TimeType",
]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import dataclasses
import typing
import abc

import pydantic
import numpy as np
import numpy.typing as npt

import spectre_core.events
import spectre_core.flowgraphs
import spectre_core.models
import spectre_core.batches
import spectre_core.spectrograms

from ._register import register_receiver
from ._base import Base, ReceiverComponents
from ._names import ReceiverName


def _is_close(
    ar: npt.NDArray[np.float32],
    ar_comparison: npt.NDArray[np.float32],
    absolute_tolerance: float,
) -> bool:
    """
    Checks if all elements in two arrays are element-wise close within a given tolerance.

    :param ar: First array for comparison.
    :param ar_comparison: Second array for comparison.
    :param absolute_tolerance: Absolute tolerance for element-wise comparison.
    :return: `True` if all elements are close within the specified tolerance, otherwise `False`.
    """
    return bool(np.all(np.isclose(ar, ar_comparison, atol=absolute_tolerance)))


M = typing.TypeVar("M", bound=pydantic.BaseModel)


class Solver(typing.Generic[M]):
    @abc.abstractmethod
    def solve(
        self, num_spectrums: int, model: M
    ) -> spectre_core.spectrograms.Spectrogram:
        """Subclasses should produce an analytically derived spectrogram for the `SignalGenerator` receiver
        according to the mode and input model.

        :param num_spectrums: The number of spectrums in the resulting spectrogram.
        :param model: The model containing the configurable parameters.
        """


class Solvers(ReceiverComponents[Solver]):
    """For each mode, produce an analytically-derived spectrogram."""


class CosineWaveSolver(Solver[spectre_core.models.SignalGeneratorCosineWaveModel]):
    def solve(
        self,
        num_spectrums: int,
        model: spectre_core.models.SignalGeneratorCosineWaveModel,
    ) -> spectre_core.spectrograms.Spectrogram:
        """Produces the analytically-derived spectrogram for the `SignalGenerator` receiver operating in the mode `cosine_wave`."""

        # Calculate derived parameters a (sampling rate ratio) and p (sampled periods).
        a = int(model.sample_rate / model.frequency)
        p = int(model.window_size / a)

        # Create the analytical spectrum, which is constant in time.
        spectrum = np.zeros(model.window_size)
        spectral_amplitude = model.amplitude * model.window_size / 2
        spectrum[p] = spectral_amplitude
        spectrum[model.window_size - p] = spectral_amplitude

        # Align spectrum to naturally ordered frequency array.
        spectrum = np.fft.fftshift(spectrum)

        # Populate the spectrogram with identical spectra.
        dynamic_spectra = (
            np.ones((model.window_size, num_spectrums)) * spectrum[:, np.newaxis]
        )

        # Compute time array.
        sampling_interval = np.float32(1 / model.sample_rate)
        times = np.arange(num_spectrums) * model.window_hop * sampling_interval

        # Compute the frequency array.
        frequencies = (
            np.fft.fftshift(np.fft.fftfreq(model.window_size, sampling_interval))
            + model.center_frequency
        )

        return spectre_core.spectrograms.Spectrogram(
            dynamic_spectra,
            times,
            frequencies,
            spectre_core.spectrograms.SpectrumUnit.AMPLITUDE,
        )


class ConstantStaircaseSolver(
    Solver[spectre_core.models.SignalGeneratorConstantStaircaseModel]
):
    def solve(
        self,
        num_spectrums: int,
        model: spectre_core.models.SignalGeneratorConstantStaircaseModel,
    ) -> spectre_core.spectrograms.Spectrogram:
        """Produces the analytically-derived spectrogram for the `SignalGenerator` receiver operating in the mode `constant_staircase`."""

        # Calculate step sizes and derived parameters.
        num_samples_per_step = np.arange(
            model.min_samples_per_step,
            model.max_samples_per_step + 1,
            model.step_increment,
        )
        num_steps = len(num_samples_per_step)

        # Create the analytical spectrum, constant in time.
        spectrum = np.zeros(model.window_size * num_steps)
        step_count = 0
        for i in range(num_steps):
            step_count += 1
            spectral_amplitude = model.window_size * step_count
            spectrum[int(model.window_size / 2) + i * model.window_size] = (
                spectral_amplitude
            )

        # Populate the spectrogram with identical spectra.
        dynamic_spectra = (
            np.ones((model.window_size * num_steps, num_spectrums))
            * spectrum[:, np.newaxis]
        )

        # Compute time array
        num_samples_per_sweep = sum(num_samples_per_step)
        sampling_interval = np.float32(1 / model.sample_rate)
        # compute the sample index we are "assigning" to each spectrum
        # and multiply by the sampling interval to get the equivalent physical time
        times = (
            np.array([(i * num_samples_per_sweep) for i in range(num_spectrums)])
            * sampling_interval
        )

        # Compute the frequency array
        baseband_frequencies = np.fft.fftshift(
            np.fft.fftfreq(model.window_size, sampling_interval)
        )
        frequencies = np.empty((model.window_size * num_steps), dtype=np.float32)
        for i in range(num_steps):
            lower_bound = i * model.window_size
            upper_bound = (i + 1) * model.window_size
            frequencies[lower_bound:upper_bound] = (
                baseband_frequencies + (model.sample_rate / 2) + (model.sample_rate * i)
            )

        # Return the spectrogram.
        return spectre_core.spectrograms.Spectrogram(
            dynamic_spectra,
            times,
            frequencies,
            spectre_core.spectrograms.SpectrumUnit.AMPLITUDE,
        )


@dataclasses.dataclass(frozen=True)
class _Mode:
    COSINE_WAVE = "cosine_wave"
    CONSTANT_STAIRCASE = "constant_staircase"


@register_receiver(ReceiverName.SIGNAL_GENERATOR)
class SignalGenerator(Base):
    def __init__(
        self, *args, solvers: typing.Optional[Solvers] = None, **kwargs
    ) -> None:
        """An entirely software-defined receiver, which generates synthetic signals."""
        super().__init__(*args, **kwargs)

        self.__solvers = solvers or Solvers()

        self.add_mode(
            _Mode.COSINE_WAVE,
            spectre_core.models.SignalGeneratorCosineWaveModel,
            spectre_core.flowgraphs.SignalGeneratorCosineWave,
            spectre_core.events.FixedCenterFrequency,
            spectre_core.batches.IQStreamBatch,
        )
        self.add_solver(_Mode.COSINE_WAVE, CosineWaveSolver())

        self.add_mode(
            _Mode.CONSTANT_STAIRCASE,
            spectre_core.models.SignalGeneratorConstantStaircaseModel,
            spectre_core.flowgraphs.SignalGeneratorConstantStaircase,
            spectre_core.events.SweptCenterFrequency,
            spectre_core.batches.IQStreamBatch,
        )
        self.add_solver(_Mode.CONSTANT_STAIRCASE, ConstantStaircaseSolver())

    @property
    def solver(
        self,
    ) -> Solver:
        return self.__solvers.get(self.active_mode)

    def add_solver(self, mode: str, solver: Solver) -> None:
        self.__solvers.add(mode, solver)

    def validate_analytically(
        self,
        spectrogram: spectre_core.spectrograms.Spectrogram,
        model: pydantic.BaseModel,
        absolute_tolerance: float,
    ) -> dict[str, typing.Any]:
        """Validate a spectrogram generated during sessions with a `SignalGenerator` receiver operating
        in a particular mode.

        :param spectrogram: The spectrogram to be validated.
        :param absolute_tolerance: Tolerance level for numerical comparisons.
        :return: A dictionary summarising the validation outcome.
        """
        analytical_spectrogram = self.solver.solve(spectrogram.num_times, model)

        # Validate times and frequencies.
        times_validated = _is_close(
            analytical_spectrogram.times, spectrogram.times, absolute_tolerance
        )
        frequencies_validated = _is_close(
            analytical_spectrogram.frequencies,
            spectrogram.frequencies,
            absolute_tolerance,
        )

        # Validate each spectrum.
        spectrum_validated = {
            spectrogram.times[i]: _is_close(
                analytical_spectrogram.dynamic_spectra[:, i],
                spectrogram.dynamic_spectra[:, i],
                absolute_tolerance,
            )
            for i in range(spectrogram.num_times)
        }

        # Summarise results
        num_validated_spectrums = sum(spectrum_validated.values())
        num_invalid_spectrums = len(spectrum_validated) - num_validated_spectrums

        return {
            "times_validated": times_validated,
            "frequencies_validated": frequencies_validated,
            "spectrum_validated": spectrum_validated,
            "num_validated_spectrums": num_validated_spectrums,
            "num_invalid_spectrums": num_invalid_spectrums,
        }



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import abc
import typing
import enum
import datetime

import numpy.typing as npt
import numpy as np
import matplotlib.axes
import matplotlib.dates
import matplotlib.figure

import spectre_core.spectrograms
import spectre_core.config
from ._format import PanelFormat
from ._panel_names import PanelName


class XAxisType(enum.Enum):
    """The xaxis type for a panel.

    Axes are shared in a stack between panels with common `XAxisType`.

    :ivar TIME: The xaxis has units of time.
    :ivar FREQUENCY: The xaxis has units of frequency.
    """

    TIME = "time"
    FREQUENCY = "frequency"


class BasePanel(abc.ABC):
    """Abstract base class for a panel used to visualise spectrogram data.

    `BasePanel` instances are designed to be part of a `PanelStack`, where multiple
    panels contribute to a composite plot. Subclasses must implement methods to define
    how the panel is drawn and annotated, and specify its xaxis type.
    """

    def __init__(
        self,
        name: PanelName,
        spectrogram: spectre_core.spectrograms.Spectrogram,
        time_type: spectre_core.spectrograms.TimeType = spectre_core.spectrograms.TimeType.RELATIVE,
    ) -> None:
        """Initialize an instance of `BasePanel`.

        :param name: The name of the panel.
        :param spectrogram: The spectrogram being visualised.
        :param time_type: Indicates whether the times of each spectrum are relative to the first
        spectrum in the spectrogram, or datetimes.
        """
        self._name = name
        self._spectrogram = spectrogram
        self._time_type = time_type

        # These attributes should be set by instances of `PanelStack`.
        self._ax: typing.Optional[matplotlib.axes.Axes] = None
        self._fig: typing.Optional[matplotlib.figure.Figure] = None
        self._panel_format: typing.Optional[PanelFormat] = None
        self._identifier: typing.Optional[str] = None

    @abc.abstractmethod
    def draw(self) -> None:
        """Modify the `ax` attribute to draw the panel contents."""

    @abc.abstractmethod
    def annotate_xaxis(self) -> None:
        """Modify the `ax` attribute to annotate the xaxis of the panel."""

    @abc.abstractmethod
    def annotate_yaxis(self) -> None:
        """Modify the `ax` attribute to annotate the yaxis of the panel."""

    @property
    @abc.abstractmethod
    def xaxis_type(self) -> XAxisType:
        """Specify the xaxis type for the panel."""

    @property
    def spectrogram(self) -> spectre_core.spectrograms.Spectrogram:
        """The spectrogram being visualised on this panel."""
        return self._spectrogram

    @property
    def name(self) -> PanelName:
        """The name of the panel."""
        return self._name

    def get_time_type(self) -> spectre_core.spectrograms.TimeType:
        """The time type of the spectrogram.

        :raises ValueError: If the `time_type` has not been set.
        """
        return self._time_type

    def set_time_type(self, value: spectre_core.spectrograms.TimeType) -> None:
        """Set the `TimeType` for the spectrogram.

        This controls how time is represented and annotated on the panel.

        :param value: The `TimeType` to assign to the spectrogram.
        """
        self._time_type = value

    def get_panel_format(self) -> PanelFormat:
        """Retrieve the panel format, which controls the style of the panel.

        :raises ValueError: If the `panel_format` has not been set.
        """
        if self._panel_format is None:
            raise ValueError(f"`panel_format` must be set for the panel `{self.name}`")
        return self._panel_format

    def set_panel_format(self, value: PanelFormat) -> None:
        """Set the panel format to control the style of the panel.

        :param value: The `PanelFormat` to assign to the panel.
        """
        self._panel_format = value

    def _get_ax(self) -> matplotlib.axes.Axes:
        """Return the `Axes` object bound to this panel.

        This method is protected to restrict direct access to `matplotlib` functionality,
        promoting encapsulation, promoting encapsulation.

        :raises ValueError: If the `Axes` object has not been set.
        """
        if self._ax is None:
            raise ValueError(f"`ax` must be set for the panel `{self.name}`")
        return self._ax

    def set_ax(self, value: matplotlib.axes.Axes) -> None:
        """Assign a Matplotlib `Axes` object to this panel.

        This `Axes` will be used for drawing and annotations.

        :param value: The Matplotlib `Axes` to assign to the panel.
        """
        self._ax = value

    def _get_fig(self) -> matplotlib.figure.Figure:
        """Return the `Figure` object bound to this panel.

        This method is protected to restrict direct access to `matplotlib` functionality,
        promoting encapsulation, promoting encapsulation.

        :raises ValueError: If the `Figure` object has not been set.
        """
        if self._fig is None:
            raise ValueError(f"`fig` must be set for the panel `{self.name}`")
        return self._fig

    def set_fig(self, value: matplotlib.figure.Figure) -> None:
        """
        Assign a Matplotlib `Figure` object to this panel.

        This `Figure` is shared across all panels in the `PanelStack`.

        :param value: The Matplotlib `Figure` to assign to the panel.
        """
        self._fig = value

    def get_identifier(self) -> typing.Optional[str]:
        """Optional identifier for the panel.

        This identifier can be used to distinguish panels or aid in superimposing
        panels in a stack.
        """
        return self._identifier

    def set_identifier(self, value: str) -> None:
        """Set the optional identifier for the panel.

        This can be used to distinguish panels or aid in superimposing panels.
        """
        self._identifier = value

    def hide_xaxis_labels(self) -> None:
        """Hide the labels for xaxis ticks in the panel."""
        self._get_ax().tick_params(axis="x", labelbottom=False)

    def hide_yaxis_labels(self) -> None:
        """Hide the labels for yaxis ticks in the panel."""
        self._get_ax().tick_params(axis="y", labelleft=False)

    def sharex(self, axes: matplotlib.axes.Axes):
        """Share the xaxis with another axes."""
        self._get_ax().sharex(axes)

    def get_xaxis_labels(self) -> list[str]:
        """Get the text string of all xaxis tick labels."""
        tick_labels = self._get_ax().get_xaxis().get_ticklabels(which="both")
        return [tick_label.get_text() for tick_label in tick_labels]

    def get_yaxis_labels(self) -> list[str]:
        """Get the text string of all yaxis tick labels."""
        tick_labels = self._get_ax().get_yaxis().get_ticklabels(which="both")
        return [tick_label.get_text() for tick_label in tick_labels]

    def get_xlabel(self) -> str:
        """Get the xlabel text string."""
        return self._get_ax().get_xlabel()

    def get_ylabel(self) -> str:
        """Get the ylabel text string."""
        return self._get_ax().get_ylabel()

    def share_axes(self, panel: "BasePanel") -> None:
        # TODO: More elegantly share axes, rather than access protected method from
        # another instance. Probably, this should just be entirely managed by the `PanelStack`
        self.set_ax(panel._get_ax())
        self.set_fig(panel._get_fig())


class BaseTimeSeriesPanel(BasePanel):
    """
    Abstract subclass of `BasePanel` designed for visualising time series data.

    Subclasses must implement any remaining abstract methods from `BasePanel`.
    """

    @property
    def xaxis_type(self) -> typing.Literal[XAxisType.TIME]:
        return XAxisType.TIME

    @property
    def times(self) -> npt.NDArray[np.float32 | np.datetime64]:
        """The times assigned to each spectrum according to the `TimeType`."""
        return (
            self.spectrogram.times
            if self.get_time_type() == spectre_core.spectrograms.TimeType.RELATIVE
            else self.spectrogram.datetimes
        )

    def annotate_xaxis(self) -> None:
        """Annotate the xaxis according to the specified `TimeType`."""
        ax = self._get_ax()
        if self.get_time_type() == spectre_core.spectrograms.TimeType.RELATIVE:
            ax.set_xlabel("Time [s]")
        else:
            # TODO: Adapt for time ranges greater than one day
            start_date = datetime.datetime.strftime(
                self.spectrogram.start_datetime.astype(datetime.datetime),
                spectre_core.config.TimeFormat.DATE,
            )
            ax.set_xlabel(f"Time [UTC] (Start Date: {start_date})")
            ax.xaxis.set_major_formatter(
                matplotlib.dates.DateFormatter(spectre_core.config.TimeFormat.TIME)
            )



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import typing

import pydantic


class Field:
    """Shared `pydantic` fields for all models."""

    max_noutput_items = typing.Annotated[
        int,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=1000,
            description="The maximum number of samples handled at each call of the work function in GNU Radio.",
        ),
    ]
    sample_rate = typing.Annotated[
        int,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=0,
            description="The number of samples taken per second, in Hz.",
        ),
    ]
    batch_size = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            ge=1,
            description="Samples are recorded in batches of this size, specified in seconds.",
        ),
    ]
    frequency = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=0,
            description="Frequency of the signal, in Hz.",
        ),
    ]
    amplitude = typing.Annotated[
        int,
        pydantic.Field(
            ...,
            validate_default=True,
            description="Amplitude of the signal, arbitrary units.",
        ),
    ]
    window_size = typing.Annotated[
        int,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=2,
            description="The size of the window, in samples, when performing the Short Time FFT.",
        ),
    ]
    window_hop = typing.Annotated[
        int,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=0,
            description="How much the window is shifted, in samples, when performing the Short Time FFT.",
        ),
    ]
    window_type = typing.Annotated[
        typing.Literal["blackman", "hann", "boxcar"],
        pydantic.Field(
            ...,
            validate_default=True,
            description="The type of window applied when performing the Short Time FFT.",
        ),
    ]
    center_frequency = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=0.0,
            description="The center frequency, in Hz. This value determines the midpoint of the frequency range being processed.",
        ),
    ]
    frequency_resolution = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            ge=0,
            description="Spectrograms are averaged up to the frequency resolution, in Hz. 0 for no averaging.",
        ),
    ]
    time_resolution = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            ge=0,
            description="Spectrograms are averaged up to the time resolution, in seconds. 0 for no averaging.",
        ),
    ]
    time_range = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            ge=0,
            description="Spectrograms are stitched together until the time range has elapsed. 0 for no stitching.",
        ),
    ]
    origin = typing.Annotated[
        str,
        pydantic.Field(
            ...,
            validate_default=True,
            description="Corresponds to the FITS keyword ORIGIN.",
        ),
    ]
    telescope = typing.Annotated[
        str,
        pydantic.Field(
            ...,
            validate_default=True,
            description="Corresponds to the FITS keyword TELESCOP.",
        ),
    ]
    instrument = typing.Annotated[
        str,
        pydantic.Field(
            ...,
            validate_default=True,
            description="Corresponds to the FITS keyword INSTRUMEN.",
        ),
    ]
    # Add an underscore to not conflict with the globally-scoped `object`.
    object_ = typing.Annotated[
        str,
        pydantic.Field(
            ...,
            validate_default=True,
            description="Corresponds to the FITS keyword OBJECT.",
        ),
    ]
    obs_lat = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            description="Corresponds to the FITS keyword OBS_LAT.",
        ),
    ]
    obs_alt = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            description="Corresponds to the FITS keyword OBS_ALT.",
        ),
    ]
    obs_lon = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            description="Corresponds to the FITS keyword OBS_LON.",
        ),
    ]
    keep_signal = typing.Annotated[
        bool,
        pydantic.Field(
            ...,
            validate_default=True,
            description="If True, keep the signal after creating the spectrogram. Otherwise, it is deleted from the file system.",
        ),
    ]
    frequency_step = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=0,
            description="The amount, in Hz, by which the center frequency is incremented for each step in the frequency sweep.",
        ),
    ]
    step_increment = typing.Annotated[
        int,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=0,
            description="The length by which each step in the staircase is incremented.",
        ),
    ]
    min_samples_per_step = typing.Annotated[
        int,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=0,
            description="The minimum number of samples in the shortest step of the staircase.",
        ),
    ]
    max_samples_per_step = typing.Annotated[
        int,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=0,
            description="The maximum number of samples in the shortest step of the staircase.",
        ),
    ]
    bandwidth = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=0,
            description="The range in Hz the signal will occupy in the frequency domain without attenutation.",
        ),
    ]
    if_gain = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            description="The intermediate frequency gain, in dB. Negative value indicates attenuation.",
        ),
    ]
    rf_gain = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            description="The radio frequency gain, in dB. Negative value indicates attenuation.",
        ),
    ]
    min_frequency = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=0,
            description="The minimum center frequency, in Hz, for the frequency sweep.",
        ),
    ]
    max_frequency = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=0,
            description="The maximum center frequency, in Hz, for the frequency sweep.",
        ),
    ]
    gain = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            description="The gain value for the SDR, in dB",
        ),
    ]
    master_clock_rate = typing.Annotated[
        int,
        pydantic.Field(
            ...,
            validate_default=True,
            description="The primary reference clock for the SDR, specified in Hz.",
        ),
    ]
    wire_format = typing.Annotated[
        str,
        pydantic.Field(
            ...,
            validate_default=True,
            description="Controls the form of the data over the bus/network.",
        ),
    ]
    samples_per_step = typing.Annotated[
        int,
        pydantic.Field(
            ...,
            validate_default=True,
            gt=0,
            description="The number of samples taken at each center frequency in the sweep. This may vary slightly from what is specified due to the nature of GNU Radio runtime.",
        ),
    ]
    antenna_port = typing.Annotated[
        str,
        pydantic.Field(
            ...,
            validate_default=True,
            description="Specifies a particular antenna port on a receiver.",
        ),
    ]
    amp_on = typing.Annotated[
        bool,
        pydantic.Field(
            ...,
            validate_default=True,
            description="If true, amplify the signal.",
        ),
    ]
    lna_gain = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            description="The low-noise amplifier gain, in dB.",
        ),
    ]
    vga_gain = typing.Annotated[
        float,
        pydantic.Field(
            ...,
            validate_default=True,
            description="The variable-gain amplifier gain, in dB.",
        ),
    ]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import typing
import collections
import datetime

import spectre_core.config
import spectre_core.spectrograms
from ._base import Base, parse_batch_file_name

T = typing.TypeVar("T", bound=Base)


class Batches(typing.Generic[T]):
    def __init__(
        self,
        tag: str,
        batch_cls: typing.Type[T],
        batches_dir_path: typing.Optional[str] = None,
    ) -> None:
        """A simple interface to read batched filesystem data.

        :param batch_cls: The `Base` subclass used to read batch files under that tag.
        :param tag: The data tag.
        :param batches_dir_path: Optionally override the directory containing the batched files.
        """
        self.__batch_cls = batch_cls
        self.__tag = tag
        self.__batches_dir_path = (
            batches_dir_path or spectre_core.config.paths.get_batches_dir_path()
        )
        self.__batch_map: dict[str, T] = collections.OrderedDict()
        self.__update()

    def __update(self) -> None:
        """Perform a fresh search of all files with `tag` in the batch name."""
        self.__batch_map.clear()

        paths = []
        for root, _, files in os.walk(self.__batches_dir_path):
            for file in files:
                paths.append(os.path.join(root, file))

        for path in paths:
            start_time, tag, _ = parse_batch_file_name(os.path.basename(path))
            if not self.__tag is None and tag == self.__tag:
                self.__batch_map[start_time] = self.__batch_cls(
                    os.path.dirname(path), start_time, tag
                )

        self.__batch_map = collections.OrderedDict(sorted(self.__batch_map.items()))

    def __iter__(self) -> typing.Iterator[T]:
        yield from self.__batch_map.values()

    def __len__(self) -> int:
        return len(self.__batch_map)

    def __getitem__(self, start_time: str) -> T:
        """Get a batch by its start time string.

        :param start_time: The start time string in the format specified by TimeFormat.DATETIME.
        :return: The batch corresponding to the start time.
        :raises KeyError: If no batch exists for the given start time.
        """
        if start_time not in self.__batch_map:
            raise KeyError(f"No batch found for start time '{start_time}'")
        return self.__batch_map[start_time]

    def __validate_range(
        self, start_datetime: datetime.datetime, end_datetime: datetime.datetime
    ) -> None:
        if start_datetime == end_datetime:
            raise ValueError(
                f"The start and end time must be different. "
                f"Got start time {start_datetime}, "
                f"and end time {end_datetime}"
            )

        if start_datetime > end_datetime:
            raise ValueError(
                f"The start time must be less than the end time. "
                f"Got start time {start_datetime}, "
                f"and end time {end_datetime}"
            )

    def get_spectrogram(
        self, start_datetime: datetime.datetime, end_datetime: datetime.datetime
    ) -> spectre_core.spectrograms.Spectrogram:
        """
        Retrieve a spectrogram spanning the specified time range.

        :param start_datetime: The start time of the range (inclusive).
        :param end_datetime: The end time of the range (inclusive).
        :raises FileNotFoundError: If no spectrogram data is available within the specified time range.
        :raise ValueError: If the start time is not less than the end time.
        :return: A spectrogram created by stitching together data from all matching batches.
        """
        self.__validate_range(start_datetime, end_datetime)
        batches_in_range = self.get_batches_in_range(start_datetime, end_datetime)
        spectrograms = [
            batch.read_spectrogram()
            for batch in batches_in_range
            if batch.spectrogram_file.exists
        ]

        if not spectrograms:
            raise FileNotFoundError(
                f"No spectrogram data found for the time range {start_datetime} to {end_datetime}."
            )
        return spectre_core.spectrograms.time_chop(
            spectre_core.spectrograms.join_spectrograms(spectrograms),
            start_datetime,
            end_datetime,
        )

    def get_batches_in_range(
        self, start_datetime: datetime.datetime, end_datetime: datetime.datetime
    ) -> list[T]:
        """Get batches that overlap with the input time range.

        The end time of each batch is assumed to be upper bounded by the start time of the next,
        since they cannot overlap. The final batch is treated as ending at `datetime.max`
        since there is no batch after it to provide that upper bound.

        :param start_datetime: The start time of the range (inclusive).
        :param end_datetime: The end time of the range (inclusive).
        :raise ValueError: If the start time is not less than the end time.
        :return: A list of batches that fall within the specified time range.
        """
        self.__validate_range(start_datetime, end_datetime)
        filtered_batches = []
        batch_datetimes = [
            datetime.datetime.strptime(t, spectre_core.config.TimeFormat.DATETIME)
            for t in self.__batch_map.keys()
        ]
        for idx, batch in enumerate(self):
            this_start = batch_datetimes[idx]
            next_start = (
                batch_datetimes[idx + 1]
                if idx + 1 < len(batch_datetimes)
                else datetime.datetime.max
            )
            if start_datetime < next_start and this_start <= end_datetime:
                filtered_batches.append(batch)

        return filtered_batches



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""Basic file io for common file formats."""

from ._files import Base, FileFormat, read_file

__all__ = ["Base", "FileFormat", "read_file"]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import dataclasses


@dataclasses.dataclass(frozen=True)
class ReceiverName:
    """The name of a supported receiver.

    :ivar CUSTOM: A customisable receiver.
    :ivar SIGNAL_GENERATOR: A synthetic signal generator.
    :ivar RSP1A: SDRPlay RSP1A.
    :ivar RSPDUO: SDRPlay RSPduo.
    :ivar RSPDX: SDRPlay RSPdx.
    :ivar USRP: A general USRP receiver.
    :ivar B200MINI: USRP B200mini.
    :ivar HACKRF: Any general HackRF receiver.
    :ivar HACKRFONE: Hack RF One.
    :ivar RTLSDR: RTL-SDR.
    """

    CUSTOM = "custom"
    SIGNAL_GENERATOR = "signal_generator"
    RSP1A = "rsp1a"
    RSPDUO = "rspduo"
    RSPDX = "rspdx"
    USRP = "usrp"
    B200MINI = "b200mini"
    HACKRF = "hackrf"
    HACKRFONE = "hackrfone"
    RTLSDR = "rtlsdr"



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import pydantic

import spectre_core.events
import spectre_core.flowgraphs

from ._validators import (
    skip_validator,
    validate_window_size,
    validate_one_of,
    validate_non_overlapping_steps,
    validate_num_samples_per_step,
    validate_nyquist_criterion,
    validate_num_steps_per_sweep,
)
from ._sdrplay_validators import (
    validate_bandwidth,
    validate_center_frequency,
    validate_if_gain,
    validate_low_if_sample_rate,
    validate_rf_gain,
    validate_sample_rate,
    validate_constant_lna_state,
)


def _get_rf_gains(center_frequency: float) -> list[int]:
    # Assume HDR mode is false.
    # Some nasty black formatting, but we move.
    if center_frequency <= 12e6:
        return [
            0,
            -3,
            -6,
            -9,
            -12,
            -15,
            -24,
            -27,
            -30,
            -33,
            -36,
            -39,
            -42,
            -45,
            -48,
            -51,
            -54,
            -57,
            -60,
        ]
    elif center_frequency <= 50e6:
        return [
            0,
            -3,
            -6,
            -9,
            -12,
            -15,
            -18,
            -24,
            -27,
            -30,
            -33,
            -36,
            -39,
            -42,
            -45,
            -48,
            -51,
            -54,
            -57,
            -60,
        ]
    elif center_frequency <= 60e6:
        return [
            0,
            -3,
            -6,
            -9,
            -12,
            -20,
            -23,
            -26,
            -29,
            -32,
            -35,
            -38,
            -44,
            -47,
            -50,
            -53,
            -56,
            -59,
            -62,
            -65,
            -68,
            -71,
            -74,
            -77,
            -80,
        ]
    elif center_frequency <= 250e6:
        return [
            0,
            -3,
            -6,
            -9,
            -12,
            -15,
            -24,
            -27,
            -30,
            -33,
            -36,
            -39,
            -42,
            -45,
            -48,
            -51,
            -54,
            -57,
            -60,
            -63,
            -66,
            -69,
            -72,
            -75,
            -78,
            -81,
            -84,
        ]
    elif center_frequency <= 420e6:
        return [
            0,
            -3,
            -6,
            -9,
            -12,
            -15,
            -18,
            -24,
            -27,
            -30,
            -33,
            -36,
            -39,
            -42,
            -45,
            -48,
            -51,
            -54,
            -57,
            -60,
            -63,
            -66,
            -69,
            -72,
            -75,
            -78,
            -81,
            -84,
        ]
    elif center_frequency <= 1000e6:
        return [
            0,
            -7,
            -10,
            -13,
            -16,
            -19,
            -22,
            -25,
            -31,
            -34,
            -37,
            -40,
            -43,
            -46,
            -49,
            -52,
            -55,
            -58,
            -61,
            -64,
            -67,
        ]
    elif center_frequency <= 2000e6:
        return [
            0,
            -5,
            -8,
            -11,
            -14,
            -17,
            -20,
            -32,
            -35,
            -38,
            -41,
            -44,
            -47,
            -50,
            -53,
            -56,
            -59,
            -62,
            -65,
        ]
    else:
        return []


class RSPdxFixedCenterFrequency(
    spectre_core.flowgraphs.RSPdxFixedCenterFrequencyModel,
    spectre_core.events.FixedCenterFrequencyModel,
):
    @pydantic.model_validator(mode="after")
    def validator(self, info: pydantic.ValidationInfo):
        if skip_validator(info):
            return self
        validate_one_of(
            self.antenna_port,
            [
                spectre_core.flowgraphs.RSPdxPort.ANT_A,
                spectre_core.flowgraphs.RSPdxPort.ANT_B,
            ],
            name="antenna_port",
        )
        validate_nyquist_criterion(self.sample_rate, self.bandwidth)
        validate_center_frequency(self.center_frequency)
        validate_window_size(self.window_size)
        validate_sample_rate(self.sample_rate)
        validate_bandwidth(self.bandwidth)
        validate_if_gain(self.if_gain)
        validate_low_if_sample_rate(self.sample_rate)
        validate_rf_gain(self.rf_gain, _get_rf_gains(self.center_frequency))
        return self


class RSPdxSweptCenterFrequency(
    spectre_core.flowgraphs.RSPdxSweptCenterFrequencyModel,
    spectre_core.events.SweptCenterFrequencyModel,
):
    @pydantic.model_validator(mode="after")
    def validator(self, info: pydantic.ValidationInfo):
        if skip_validator(info):
            return self
        validate_one_of(
            self.antenna_port,
            [
                spectre_core.flowgraphs.RSPdxPort.ANT_A,
                spectre_core.flowgraphs.RSPdxPort.ANT_B,
            ],
            name="antenna_port",
        )
        validate_center_frequency(self.min_frequency)
        validate_center_frequency(self.max_frequency)
        validate_window_size(self.window_size)
        validate_sample_rate(self.sample_rate)
        validate_bandwidth(self.bandwidth)
        validate_if_gain(self.if_gain)
        validate_low_if_sample_rate(self.sample_rate)
        validate_non_overlapping_steps(self.frequency_step, self.sample_rate)
        validate_num_samples_per_step(self.window_size, self.samples_per_step)
        validate_num_steps_per_sweep(
            self.min_frequency, self.max_frequency, self.frequency_step
        )
        validate_constant_lna_state(
            self.min_frequency, self.max_frequency, _get_rf_gains
        )
        return self



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import enum

import numpy as np
import numpy.typing as npt
import pyfftw


def get_cosine_signal(
    num_samples: int,
    sample_rate: float,
    frequency: float,
    amplitude: float,
    phase: float,
) -> npt.NDArray[np.complex64]:
    """Generate a complex-valued cosine wave.

    :param num_samples: The number of samples in the signal.
    :param sample_rate: The rate at which the signal is sampled.
    :param frequency: The frequency of the cosine.
    :param amplitude: The amplitude of the cosine.
    :param phase: The initial phase of the cosine.
    :return: A numpy array containing the complex-valued signal.
    """
    n = np.arange(num_samples, dtype=np.complex64)
    return amplitude * np.cos(2 * np.pi * (frequency / sample_rate) * n + phase)


class WindowType(str, enum.Enum):
    HANN = "hann"
    BLACKMAN = "blackman"
    BOXCAR = "boxcar"


def _window_general_cosine_asym(
    window_size: int,
    coefficients: npt.NDArray[np.float64],
) -> npt.NDArray[np.float32]:
    angles = np.linspace(-np.pi, np.pi, window_size + 1)[:-1]
    window = sum(coef * np.cos(k * angles) for k, coef in enumerate(coefficients))
    return window.astype(np.float32)


def _window_boxcar(window_size: int) -> npt.NDArray[np.float32]:
    return np.ones(window_size, np.float32)


def _window_hann(window_size: int) -> npt.NDArray[np.float32]:
    coefficients = np.asarray([0.5, 0.5])
    return _window_general_cosine_asym(window_size, coefficients)


def _window_blackman(window_size: int) -> npt.NDArray[np.float32]:
    coefficients = np.asarray([0.42, 0.50, 0.08])
    return _window_general_cosine_asym(window_size, coefficients)


def get_window(window_type: WindowType, window_size: int) -> npt.NDArray[np.float32]:
    """Create a window of a specified type and length.

    :param window_type: The type of window to generate.
    :param window_size: The number of samples in the window.
    :return: A numpy array containing the window samples.
    :raises ValueError: If window_size is less than two or an unknown window type is provided.
    """
    if window_size < 2:
        raise ValueError(f"The window size cannot be less than 2, got {window_size}")

    if window_type == WindowType.BOXCAR:
        return _window_boxcar(window_size)
    elif window_type == WindowType.HANN:
        return _window_hann(window_size)
    elif window_type == WindowType.BLACKMAN:
        return _window_blackman(window_size)
    else:
        raise ValueError(f"Unknown window type: {window_type}")


def get_buffer(num_samples: int) -> npt.NDArray[np.complex64]:
    """Create an empty, memory-aligned buffer for in-place DFTs carried out by FFTW.

    :param num_samples: The number of samples in the buffer.
    :return: An empty numpy array.
    """
    return pyfftw.empty_aligned(num_samples, dtype="complex64")


def get_fftw_obj(buffer: npt.NDArray[np.complex64]) -> pyfftw.FFTW:
    """Plan an in-place 1D DFT using FFTW.

    The contents of the input buffer will be overwritten during the planning process, and so
    should be initialised after this function is called.

    :param buffer: An empty numpy array.
    :return: An FFTW object that, when called, computes the forward FFT of whatever is in the buffer.
    """
    return pyfftw.FFTW(buffer, buffer, direction="FFTW_FORWARD", flags=["FFTW_PATIENT"])


def get_times(
    num_spectrums: int, sample_rate: float, window_hop: int
) -> npt.NDArray[np.float32]:
    """Get the physical times assigned to each spectrum in a spectrogram.

    The first spectrum is by convention at t=0. Each spectrum thereafter is offset by `window_hop` samples,
    corresponding to `window_hop * sample_interval` units, where `sample_interval` is the inverse of the
    sample rate.

    :param num_spectrums: The number of spectrums in the spectrogram.
    :param sample_rate: The sample rate of the signal.
    :param window_hop: The number of samples the window shifts per frame.
    :return: A numpy array containing the times assigned to each spectrum.
    """
    return np.arange(num_spectrums, dtype=np.float32) * window_hop * 1.0 / sample_rate


def get_frequencies(window_size: int, sample_rate: float) -> npt.NDArray[np.float32]:
    """Compute the DFT sample frequencies. for a given window size and sample rate.

    :param window_size: The number of samples in each window.
    :param sample_rate: The sample rate at which the signal was captured.
    :return: A numpy array containing the sample frequencies, with zero at the start.
    """
    return np.fft.fftfreq(window_size, d=np.float32(1.0 / sample_rate))


def get_num_spectrums(signal_size: int, window_size: int, window_hop: int) -> int:
    """Compute the number of spectrums in the spectrogram.

    The first window is centered at the start of the signal (index 0). The last window is the final one that
    fits entirely within the signal.

    :param signal_size: The total number of samples in the signal.
    :param window_size: The number of samples in each window.
    :param window_hop: The number of samples the window is shifted in each frame.
    :return: The total number of spectrums in the resulting spectrogram, when performing an
    stfft with these values.
    :raises ValueError: If the window size or hop is less than one sample, or the window size is greater than the signal size.
    """
    if window_size < 1:
        raise ValueError(
            f"The window size must be at least one. " f"Got {window_size}."
        )

    if window_hop < 1:
        raise ValueError(f"The window hop must be at least one. " f"Got {window_hop}.")

    if window_size > signal_size:
        raise ValueError(
            f"The window must fit within the signal. "
            f"Got window size {window_size}, which is greater "
            f"than the signal size {signal_size}."
        )

    return int((signal_size - np.ceil(window_size / 2)) / window_hop) + 1


def stfft(
    fftw_obj: pyfftw.FFTW,
    buffer: npt.NDArray[np.complex64],
    signal: npt.NDArray[np.complex64],
    window: npt.NDArray[np.float32],
    window_hop: int,
) -> npt.NDArray[np.float32]:
    """Compute the short-time discrete Fourier transform of the input signal, using a real sliding window.

    The first window is centered at the start of the signal (index 0). The last window is the final one that
    fits entirely within the signal.

    :param fftw_obj: An FFTW object, pre-planned for in-place transforms on the buffer.
    :param buffer: An empty numpy array, used for repeated in-place DFTs.
    :param signal: The input signal.
    :param window: The window function, same length as the buffer.
    :param window_hop: The number of samples the window advances per frame.
    :param sample_rate: The sample rate of the signal.
    :return: a spectrogram containing the amplitude of each spectral component.
    :raises ValueError: If the window and buffer sizes do not match.
    """
    window_size = window.shape[0]
    buffer_size = buffer.shape[0]
    signal_size = signal.shape[0]

    if window_size != buffer_size:
        raise ValueError(
            f"The window and buffer must be the same size. "
            f"Got that the window has {window_size} samples, "
            f"but the buffer has {buffer_size} samples."
        )

    # Calculate how many spectrums will be in the spectrogram.
    num_spectrums = get_num_spectrums(signal_size, window_size, window_hop)

    # Initialise an empty array, into which we'll copy the spectrums computed by fftw.
    dynamic_spectra = np.empty((window_size, num_spectrums), dtype=np.float32)

    for n in range(num_spectrums):
        # Center the window for the current frame
        center = window_hop * n
        start = center - window_size // 2
        stop = start + window_size

        # The window is fully inside the signal.
        if start >= 0 and stop <= signal_size:
            buffer[:] = signal[start:stop] * window

        # The window partially overlaps with the signal.
        else:
            # Zero the buffer and apply the window only to valid signal samples
            signal_indices = np.arange(start, stop)
            valid_mask = (signal_indices >= 0) & (signal_indices < signal_size)
            buffer[:] = 0.0
            buffer[valid_mask] = signal[signal_indices[valid_mask]] * window[valid_mask]

        # Compute the DFT in-place, to produce the spectrum.
        fftw_obj.execute()

        # Copy the spectrum into the spectrogram.
        dynamic_spectra[:, n] = np.abs(buffer)

    return dynamic_spectra



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import dataclasses

from gnuradio import sdrplay3
from gnuradio import spectre

import spectre_core.fields

from ._base import Base, BaseModel


@dataclasses.dataclass(frozen=True)
class RSPdxPort:
    """Specifies one of the antenna ports on the RSPdx.

    These are easier to type in the CLI tool, than the values we must pass to `gr-sdrplay3`.
    Namely 'Antenna A', 'Antenna B' and 'Antenna C'
    """

    ANT_A = "ant_a"
    ANT_B = "ant_b"


def _map_port(antenna_port: str | None) -> str:
    """Maps the CLI-typeable port, to the one used in the constructor for the RSPdx block in the gr-sdrplay3 OOT module."""
    # Add a typing check for None, to satisfy mypy
    if antenna_port == RSPdxPort.ANT_A:
        return "Antenna A"
    elif antenna_port == RSPdxPort.ANT_B:
        return "Antenna B"
    else:
        raise ValueError(f"{antenna_port} is not a valid antenna port.")


class RSPdxFixedCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 500000
    batch_size: spectre_core.fields.Field.batch_size = 3
    center_frequency: spectre_core.fields.Field.center_frequency = 95.8e6
    bandwidth: spectre_core.fields.Field.bandwidth = 300000
    if_gain: spectre_core.fields.Field.if_gain = -30
    rf_gain: spectre_core.fields.Field.rf_gain = 0
    antenna_port: spectre_core.fields.Field.antenna_port = RSPdxPort.ANT_A


class RSPdxFixedCenterFrequency(Base[RSPdxFixedCenterFrequencyModel]):
    def configure(self, tag: str, model: RSPdxFixedCenterFrequencyModel) -> None:
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path, tag, model.batch_size, model.sample_rate
        )

        self.sdrplay3_rspdx = sdrplay3.rspdx(
            "",
            stream_args=sdrplay3.stream_args(output_type="fc32", channels_size=1),
        )
        self.sdrplay3_rspdx.set_sample_rate(model.sample_rate)
        self.sdrplay3_rspdx.set_center_freq(model.center_frequency)
        self.sdrplay3_rspdx.set_bandwidth(model.bandwidth)
        self.sdrplay3_rspdx.set_antenna(_map_port(model.antenna_port))
        self.sdrplay3_rspdx.set_gain_mode(False)
        self.sdrplay3_rspdx.set_gain(model.if_gain, "IF")
        self.sdrplay3_rspdx.set_gain(model.rf_gain, "RF", False)
        self.sdrplay3_rspdx.set_freq_corr(0)
        self.sdrplay3_rspdx.set_dc_offset_mode(False)
        self.sdrplay3_rspdx.set_iq_balance_mode(False)
        self.sdrplay3_rspdx.set_agc_setpoint(-30)
        self.sdrplay3_rspdx.set_hdr_mode(False)
        self.sdrplay3_rspdx.set_rf_notch_filter(False)
        self.sdrplay3_rspdx.set_dab_notch_filter(False)
        self.sdrplay3_rspdx.set_biasT(False)
        self.sdrplay3_rspdx.set_stream_tags(False)
        self.sdrplay3_rspdx.set_debug_mode(False)
        self.sdrplay3_rspdx.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rspdx.set_show_gain_changes(False)

        self.connect((self.sdrplay3_rspdx, 0), (self.spectre_batched_file_sink, 0))


class RSPdxSweptCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 2000000
    batch_size: spectre_core.fields.Field.batch_size = 3
    bandwidth: spectre_core.fields.Field.bandwidth = 1.536e6
    if_gain: spectre_core.fields.Field.if_gain = -30
    rf_gain: spectre_core.fields.Field.rf_gain = 0
    min_frequency: spectre_core.fields.Field.min_frequency = 95e6
    max_frequency: spectre_core.fields.Field.max_frequency = 100e6
    samples_per_step: spectre_core.fields.Field.samples_per_step = 120000
    frequency_step: spectre_core.fields.Field.frequency_step = 2e6
    antenna_port: spectre_core.fields.Field.antenna_port = RSPdxPort.ANT_A


class RSPdxSweptCenterFrequency(Base[RSPdxSweptCenterFrequencyModel]):
    def configure(self, tag: str, model: RSPdxSweptCenterFrequencyModel) -> None:
        self.spectre_sweep_driver = spectre.sweep_driver(
            model.min_frequency,
            model.max_frequency,
            model.frequency_step,
            model.sample_rate,
            model.samples_per_step,
            "freq",
        )
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.batch_size,
            model.sample_rate,
            True,
            "freq",
            model.min_frequency,
        )
        self.sdrplay3_rspdx = sdrplay3.rspdx(
            "",
            stream_args=sdrplay3.stream_args(output_type="fc32", channels_size=1),
        )
        self.sdrplay3_rspdx.set_sample_rate(model.sample_rate)
        self.sdrplay3_rspdx.set_center_freq(model.min_frequency)
        self.sdrplay3_rspdx.set_bandwidth(model.bandwidth)
        self.sdrplay3_rspdx.set_antenna(_map_port(model.antenna_port))
        self.sdrplay3_rspdx.set_gain_mode(False)
        self.sdrplay3_rspdx.set_gain(model.if_gain, "IF")
        self.sdrplay3_rspdx.set_gain(model.rf_gain, "RF", False)
        self.sdrplay3_rspdx.set_freq_corr(0)
        self.sdrplay3_rspdx.set_dc_offset_mode(False)
        self.sdrplay3_rspdx.set_iq_balance_mode(False)
        self.sdrplay3_rspdx.set_agc_setpoint(-30)
        self.sdrplay3_rspdx.set_hdr_mode(False)
        self.sdrplay3_rspdx.set_rf_notch_filter(False)
        self.sdrplay3_rspdx.set_dab_notch_filter(False)
        self.sdrplay3_rspdx.set_biasT(False)
        self.sdrplay3_rspdx.set_stream_tags(True)
        self.sdrplay3_rspdx.set_debug_mode(False)
        self.sdrplay3_rspdx.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rspdx.set_show_gain_changes(False)

        self.msg_connect(
            (self.spectre_sweep_driver, "retune_command"),
            (self.sdrplay3_rspdx, "command"),
        )
        self.connect((self.sdrplay3_rspdx, 0), (self.spectre_batched_file_sink, 0))
        self.connect((self.sdrplay3_rspdx, 0), (self.spectre_sweep_driver, 0))



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from enum import Enum


class PanelName(Enum):
    """Literal corresponding to a fully implemented `BasePanel` subclass.

    :ivar SPECTROGRAM: Panel for visualising the full spectrogram.
    :ivar FREQUENCY_CUTS: Panel for visualising individual spectrums in a spectrogram.
    :ivar TIME_CUTS: Panel for visualising spectrogram data as time series of spectral components.
    :ivar INTEGRAL_OVER_FREQUENCY: Panel for visualising the spectrogram integrated over frequency.
    """

    SPECTROGRAM = "spectrogram"
    FREQUENCY_CUTS = "frequency_cuts"
    TIME_CUTS = "time_cuts"
    INTEGRAL_OVER_FREQUENCY = "integral_over_frequency"



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from gnuradio import sdrplay3
from gnuradio import spectre

import spectre_core.fields

from ._base import Base, BaseModel


class RSP1AFixedCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 500000
    batch_size: spectre_core.fields.Field.batch_size = 3
    center_frequency: spectre_core.fields.Field.center_frequency = 95.8e6
    bandwidth: spectre_core.fields.Field.bandwidth = 300000
    if_gain: spectre_core.fields.Field.if_gain = -30
    rf_gain: spectre_core.fields.Field.rf_gain = 0


class RSP1AFixedCenterFrequency(Base[RSP1AFixedCenterFrequencyModel]):
    def configure(self, tag: str, model: RSP1AFixedCenterFrequencyModel) -> None:
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path, tag, model.batch_size, model.sample_rate
        )
        self.sdrplay3_rsp1a = sdrplay3.rsp1a(
            "",
            stream_args=sdrplay3.stream_args(output_type="fc32", channels_size=1),
        )
        self.sdrplay3_rsp1a.set_sample_rate(model.sample_rate)
        self.sdrplay3_rsp1a.set_center_freq(model.center_frequency)
        self.sdrplay3_rsp1a.set_bandwidth(model.bandwidth)
        self.sdrplay3_rsp1a.set_gain_mode(False)
        self.sdrplay3_rsp1a.set_gain(model.if_gain, "IF")
        self.sdrplay3_rsp1a.set_gain(model.rf_gain, "RF")
        self.sdrplay3_rsp1a.set_freq_corr(0)
        self.sdrplay3_rsp1a.set_dc_offset_mode(False)
        self.sdrplay3_rsp1a.set_iq_balance_mode(False)
        self.sdrplay3_rsp1a.set_agc_setpoint(-30)
        self.sdrplay3_rsp1a.set_rf_notch_filter(False)
        self.sdrplay3_rsp1a.set_dab_notch_filter(False)
        self.sdrplay3_rsp1a.set_biasT(False)
        self.sdrplay3_rsp1a.set_debug_mode(False)
        self.sdrplay3_rsp1a.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rsp1a.set_show_gain_changes(False)

        self.connect((self.sdrplay3_rsp1a, 0), (self.spectre_batched_file_sink, 0))


class RSP1ASweptCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 2000000
    batch_size: spectre_core.fields.Field.batch_size = 3
    bandwidth: spectre_core.fields.Field.bandwidth = 1.536e6
    if_gain: spectre_core.fields.Field.if_gain = -30
    rf_gain: spectre_core.fields.Field.rf_gain = 0
    min_frequency: spectre_core.fields.Field.min_frequency = 95e6
    max_frequency: spectre_core.fields.Field.max_frequency = 100e6
    samples_per_step: spectre_core.fields.Field.samples_per_step = 120000
    frequency_step: spectre_core.fields.Field.frequency_step = 2e6


class RSP1ASweptCenterFrequency(Base[RSP1ASweptCenterFrequencyModel]):
    def configure(self, tag: str, model: RSP1ASweptCenterFrequencyModel) -> None:
        self.spectre_sweep_driver = spectre.sweep_driver(
            model.min_frequency,
            model.max_frequency,
            model.frequency_step,
            model.sample_rate,
            model.samples_per_step,
            "freq",
        )
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.batch_size,
            model.sample_rate,
            True,
            "freq",
            model.min_frequency,
        )
        self.sdrplay3_rsp1a = sdrplay3.rsp1a(
            "",
            stream_args=sdrplay3.stream_args(output_type="fc32", channels_size=1),
        )
        self.sdrplay3_rsp1a.set_sample_rate(model.sample_rate, True)
        self.sdrplay3_rsp1a.set_center_freq(model.min_frequency, True)
        self.sdrplay3_rsp1a.set_bandwidth(model.bandwidth)
        self.sdrplay3_rsp1a.set_gain_mode(False)
        self.sdrplay3_rsp1a.set_gain(model.if_gain, "IF", True)
        self.sdrplay3_rsp1a.set_gain(model.rf_gain, "RF", True)
        self.sdrplay3_rsp1a.set_freq_corr(0)
        self.sdrplay3_rsp1a.set_dc_offset_mode(False)
        self.sdrplay3_rsp1a.set_iq_balance_mode(False)
        self.sdrplay3_rsp1a.set_agc_setpoint(-30)
        self.sdrplay3_rsp1a.set_rf_notch_filter(False)
        self.sdrplay3_rsp1a.set_dab_notch_filter(False)
        self.sdrplay3_rsp1a.set_biasT(False)
        self.sdrplay3_rsp1a.set_stream_tags(True)
        self.sdrplay3_rsp1a.set_debug_mode(False)
        self.sdrplay3_rsp1a.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rsp1a.set_show_gain_changes(False)

        self.msg_connect(
            (self.spectre_sweep_driver, "retune_command"),
            (self.sdrplay3_rsp1a, "command"),
        )
        self.connect((self.sdrplay3_rsp1a, 0), (self.spectre_batched_file_sink, 0))
        self.connect((self.sdrplay3_rsp1a, 0), (self.spectre_sweep_driver, 0))



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""Manage background tasks and long-running processes."""

from ._jobs import Job, start_job
from ._workers import Worker, make_worker
from ._duration import Duration

__all__ = ["Job", "Worker", "make_worker", "start_job", "Duration"]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime
import typing
import functools
import abc
import dataclasses
import os

import spectre_core.io
import spectre_core.config
import spectre_core.spectrograms


def parse_batch_file_path(absolute_file_path: str) -> tuple[str, str, str, str]:
    """Parse a file path into the directory name, the start time, tag and extension."""
    return (
        os.path.dirname(absolute_file_path),
        *parse_batch_file_name(os.path.basename(absolute_file_path)),
    )


def parse_batch_file_name(file_name: str) -> tuple[str, str, str]:
    """Parse a file name into a start time, tag, and extension."""
    batch_name, extension = os.path.splitext(file_name)
    if batch_name.count("_") != 1:
        raise ValueError(f"Expected exactly one underscore in '{batch_name}'.")

    start_time, tag = batch_name.split("_", 1)
    return start_time, tag, extension.lstrip(".")


T = typing.TypeVar("T")


class BatchFile(spectre_core.io.Base[T]):
    def __init__(self, file_path: str) -> None:
        """An abstract base class for files belonging to a batch, identified by their file extension.

        Batch file names must conform to the following structure:

            `<start time>_<tag>.<extension>`

        The substring `<start time>_<tag>` is referred to as the batch name. Files with the same batch name
        belong to the same batch.

        :param file_path: The absolute path to the batch file.
        """
        super().__init__(file_path)
        self._start_time, self._tag, _ = parse_batch_file_name(self.file_name)

    @property
    def start_time(self) -> str:
        """The start time of the batch, up to seconds precision."""
        return self._start_time

    @property
    def tag(self) -> str:
        """The data tag."""
        return self._tag

    @functools.cached_property
    def start_datetime(self) -> datetime.datetime:
        """The start time of the batch, up to seconds precision."""
        return datetime.datetime.strptime(
            self.start_time, spectre_core.config.TimeFormat.DATETIME
        )


@dataclasses.dataclass(frozen=True)
class _Extension:
    PNG = "png"


class _PngFile(BatchFile[str]):
    """Stores an image in the PNG file format."""

    def read(self) -> str:
        return spectre_core.io.read_file(self.file_path, spectre_core.io.FileFormat.PNG)


class Base(abc.ABC):
    def __init__(self, batches_dir_path: str, start_time: str, tag: str) -> None:
        """An abstract base class representing a group of data files over a common time interval.

        All files in a batch share a base file name and differ only by their extension.
        Subclasses define the expected data for each file extension and
        provide an API for accessing their contents using `BatchFile` subclasses.

        Subclasses should expose `BatchFile` instances directly as attributes, which
        simplifies static typing. Additionally, they should call `add_file` in the constructor
        to formally register each `BatchFile`.

        :param batches_dir_path: The shared parent directory for each batch file.
        :param start_time: The start time of the batch.
        :param tag: The data tag.
        """
        self._batches_dir_path = batches_dir_path
        self._start_time = start_time
        self._tag: str = tag

        self._batch_files: dict[str, BatchFile] = {}

        self.add_file(_PngFile, _Extension.PNG)

    @property
    @abc.abstractmethod
    def spectrogram_file(self) -> BatchFile[spectre_core.spectrograms.Spectrogram]:
        """Indicate the file in the batch storing spectrogram data."""

    @property
    def start_time(self) -> str:
        """The start time of the batch, up to seconds precision."""
        return self._start_time

    @functools.cached_property
    def start_datetime(self) -> datetime.datetime:
        """The start time of the batch, up to seconds precision."""
        return datetime.datetime.strptime(
            self.start_time, spectre_core.config.TimeFormat.DATETIME
        )

    @property
    def tag(self) -> str:
        """The data tag."""
        return self._tag

    @property
    def name(self) -> str:
        """Return the base file name shared by all files in the batch,
        composed of the start time and the batch tag."""
        return f"{self._start_time}_{self._tag}"

    @property
    def extensions(self) -> list[str]:
        """All defined file extensions for the batch."""
        return list(self._batch_files.keys())

    @property
    def png_file(self) -> _PngFile:
        """The batch file corresponding to the `.png` extension."""
        return typing.cast(_PngFile, self.get_file(_Extension.PNG))

    def add_file(self, batch_file_cls: typing.Type[BatchFile], extension: str) -> None:
        """Add a batch file to the batch."""
        if extension in self._batch_files:
            raise ValueError(
                f"A file with extension '{extension}' is already registered."
            )
        self._batch_files[extension] = batch_file_cls(
            os.path.join(self._batches_dir_path, f"{self.name}.{extension}")
        )

    def get_file(self, extension: str) -> BatchFile:
        """Get a batch file from the batch, according to the file extension.

        :param extension: The file extension of the batch file.
        :raises NotImplementedError: If the extension is undefined for the batch.
        :return: The batch file registered under the input file extension.
        """
        if extension in self._batch_files:
            return self._batch_files[extension]
        else:
            raise NotImplementedError(
                f"A batch file with extension '{extension}' is not implemented for this batch."
            )

    def delete_file(self, extension: str) -> None:
        """Delete a file from the batch, according to the file extension.

        :param extension: The file extension of the batch file.
        :raises FileNotFoundError: If the batch file does not exist in the file system.
        """
        batch_file = self.get_file(extension)
        batch_file.delete()

    def has_file(self, extension: str) -> bool:
        """Determine the existance of a batch file in the file system.

        :param extension: The file extension of the batch file.
        :return: True if the batch file exists in the file system, False otherwise.
        """
        try:
            batch_file = self.get_file(extension)
            return batch_file.exists
        except FileNotFoundError:
            return False

    def read_spectrogram(self) -> spectre_core.spectrograms.Spectrogram:
        """Read and return the spectrogram data stored in the batch.

        :return: The spectrogram stored by the batch `spectrogram_file`.
        """
        return self.spectrogram_file.read()



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import typing

from ._base import Base

# map populated at runtime via the `register_receiver` decorator.
receivers: dict[str, typing.Type[Base]] = {}

T = typing.TypeVar("T", bound=Base)


def register_receiver(
    receiver_name: str,
) -> typing.Callable[[typing.Type[T]], typing.Type[T]]:
    """Decorator to register a fully implemented `Base` subclass under a specified `receiver_name`.

    :param receiver_name: The name of the receiver.
    :raises ValueError: If the provided `receiver_name` is already registered.
    :return: A decorator that registers the `Base` subclass under the given `receiver_name`.
    """

    def decorator(cls: typing.Type[T]) -> typing.Type[T]:
        if receiver_name in receivers:
            raise ValueError(f"The receiver '{receiver_name}' is already registered!")
        receivers[receiver_name] = cls
        return cls

    return decorator


def get_registered_receivers() -> list[str]:
    """List all registered receivers."""
    return [k for k in receivers.keys()]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later


class Duration:
    """A selection of set durations, in seconds."""

    ONE_CENTISECOND = 0.01
    ONE_DECISECOND = 0.1
    ONE_SECOND = 1.0
    TEN_SECONDS = 10.0



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import pydantic

import spectre_core.events
import spectre_core.flowgraphs

from ._validators import (
    skip_validator,
    validate_in_range,
    validate_window_size,
    validate_nyquist_criterion,
    validate_non_overlapping_steps,
    validate_num_steps_per_sweep,
    validate_num_samples_per_step,
)
from ._usrp_validators import (
    validate_wire_format,
    validate_sample_rate_with_master_clock_rate,
)

SAMPLE_RATE_LOWER_BOUND = 200000
SAMPLE_RATE_UPPER_BOUND = 56000000
CENTER_FREQUENCY_LOWER_BOUND = 70000000
CENTER_FREQUENCY_UPPER_BOUND = 6000000000
BANDWIDTH_LOWER_BOUND = 200000
BANDWIDTH_UPPER_BOUND = 56000000
GAIN_LOWER_BOUND = 0
GAIN_UPPER_BOUND = 76
MASTER_CLOCK_RATE_LOWER_BOUND = 5000000
MASTER_CLOCK_RATE_UPPER_BOUND = 61440000


class B200miniFixedCenterFrequency(
    spectre_core.flowgraphs.USRPFixedCenterFrequencyModel,
    spectre_core.events.FixedCenterFrequencyModel,
):
    @pydantic.model_validator(mode="after")
    def validator(self, info: pydantic.ValidationInfo):
        if skip_validator(info):
            return self
        validate_in_range(
            self.center_frequency,
            lower_bound=CENTER_FREQUENCY_LOWER_BOUND,
            upper_bound=CENTER_FREQUENCY_UPPER_BOUND,
            name="center_frequency",
        )
        validate_in_range(
            self.bandwidth,
            lower_bound=BANDWIDTH_LOWER_BOUND,
            upper_bound=BANDWIDTH_UPPER_BOUND,
            name="bandwidth",
        )
        validate_in_range(
            self.sample_rate,
            lower_bound=SAMPLE_RATE_LOWER_BOUND,
            upper_bound=SAMPLE_RATE_UPPER_BOUND,
            name="sample_rate",
        )
        validate_in_range(
            self.gain,
            lower_bound=GAIN_LOWER_BOUND,
            upper_bound=GAIN_UPPER_BOUND,
            name="gain",
        )
        validate_in_range(
            self.master_clock_rate,
            lower_bound=MASTER_CLOCK_RATE_LOWER_BOUND,
            upper_bound=MASTER_CLOCK_RATE_UPPER_BOUND,
            name="master_clock_rate",
        )
        validate_nyquist_criterion(self.sample_rate, self.bandwidth)
        validate_window_size(self.window_size)
        validate_wire_format(self.wire_format)
        validate_sample_rate_with_master_clock_rate(
            self.sample_rate, self.master_clock_rate
        )
        return self


class B200miniSweptCenterFrequency(
    spectre_core.flowgraphs.USRPSweptCenterFrequencyModel,
    spectre_core.events.SweptCenterFrequencyModel,
):
    @pydantic.model_validator(mode="after")
    def validator(self, info: pydantic.ValidationInfo):
        if skip_validator(info):
            return self
        validate_in_range(
            self.min_frequency,
            lower_bound=CENTER_FREQUENCY_LOWER_BOUND,
            upper_bound=CENTER_FREQUENCY_UPPER_BOUND,
            name="center_frequency",
        )
        validate_in_range(
            self.max_frequency,
            lower_bound=CENTER_FREQUENCY_LOWER_BOUND,
            upper_bound=CENTER_FREQUENCY_UPPER_BOUND,
            name="center_frequency",
        )
        validate_in_range(
            self.bandwidth,
            lower_bound=BANDWIDTH_LOWER_BOUND,
            upper_bound=BANDWIDTH_UPPER_BOUND,
            name="bandwidth",
        )
        validate_in_range(
            self.sample_rate,
            lower_bound=SAMPLE_RATE_LOWER_BOUND,
            upper_bound=SAMPLE_RATE_UPPER_BOUND,
            name="sample_rate",
        )
        validate_in_range(
            self.gain,
            lower_bound=GAIN_LOWER_BOUND,
            upper_bound=GAIN_UPPER_BOUND,
            name="gain",
        )
        validate_in_range(
            self.master_clock_rate,
            lower_bound=MASTER_CLOCK_RATE_LOWER_BOUND,
            upper_bound=MASTER_CLOCK_RATE_UPPER_BOUND,
            name="master_clock_rate",
        )
        validate_window_size(self.window_size)
        validate_non_overlapping_steps(self.frequency_step, self.sample_rate)
        validate_num_samples_per_step(self.window_size, self.samples_per_step)
        validate_num_steps_per_sweep(
            self.min_frequency, self.max_frequency, self.frequency_step
        )
        validate_sample_rate_with_master_clock_rate(
            self.sample_rate, self.master_clock_rate
        )
        return self



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass


@dataclass(frozen=True)
class TimeFormat:
    """Package-wide datetime formats.

    :ivar DATE: Format for dates (e.g., '2025-01-11').
    :ivar TIME: Format for times (e.g., '23:59:59').
    :ivar DATETIME: Combined date and time format (e.g., '2025-01-11T23:59:59').
    :ivar PRECISE_TIME: Format for times with microseconds (e.g., '23:59:59.123456').
    :ivar PRECISE_DATETIME: Combined date and precise time format (e.g., '2025-01-11T23:59:59.123456').
    """

    DATE = "%Y-%m-%d"
    TIME = "%H:%M:%S"
    DATETIME = f"{DATE}T{TIME}"
    PRECISE_TIME = "%H:%M:%S.%f"
    PRECISE_DATETIME = f"{DATE}T{PRECISE_TIME}"
